"""Implementation of context provider functionality."""
from __future__ import annotations

import time
import uuid
from collections import defaultdict
from typing import TYPE_CHECKING

from sdc11073.provider.operations import ExecuteResult

from . import providerbase

if TYPE_CHECKING:
    from lxml import etree

    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase

    from .providerbase import OperationClassGetter


class GenericContextProvider(providerbase.ProviderRole):
    """Handles SetContextState operations."""

    def __init__(self, mdib: ProviderMdibProtocol,
                 op_target_descr_types: list[etree.QName] | None = None,
                 log_prefix: str | None = None):
        super().__init__(mdib, log_prefix)
        self._op_target_descr_types = op_target_descr_types

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Create an OperationDefinition for SetContextStateOperationDescriptor.

        Only if type of operation target matches opTargetDescriptorTypes.
        """
        pm_names = self._mdib.data_model.pm_names
        if pm_names.SetContextStateOperationDescriptor == operation_descriptor_container.NODETYPE:
            op_target_entity = self._mdib.entities.by_handle( operation_descriptor_container.OperationTarget)

            if (not self._op_target_descr_types) or (
                    op_target_entity.descriptor.NODETYPE not in self._op_target_descr_types):
                return None  # we do not handle this target type
            return self._mk_operation_from_operation_descriptor(operation_descriptor_container,
                                                                operation_cls_getter,
                                                                operation_handler=self._set_context_state)
        return None

    def _set_context_state(self, params: ExecuteParameters) -> ExecuteResult: # noqa: C901, PLR0912
        """Execute the operation itself (ExecuteHandler).

        If the proposed context is a new context and ContextAssociation == pm_types.ContextAssociation.ASSOCIATED,
        the before associates state(s) will be set to DISASSOCIATED and UnbindingMdibVersion/BindingEndTime
        are set.
        """
        pm_types = self._mdib.data_model.pm_types
        proposed_context_states = params.operation_request.argument

        # check if there is more than one associated state for a context descriptor in proposed_context_states
        proposed_by_handle = defaultdict(list)
        for st in proposed_context_states:
            if st.ContextAssociation == pm_types.ContextAssociation.ASSOCIATED:
                proposed_by_handle[st.DescriptorHandle].append(st)
        for handle, states in proposed_by_handle.items():
            if len(states) > 1:
                msg = f'more than one associated context for descriptor handle {handle}'
                raise ValueError(msg)

        operation_target_handles = []
        modified_state_handles: dict[str, list[str]] = defaultdict(list)
        modified_entities = []
        with self._mdib.context_state_transaction() as mgr:
            for proposed_st in proposed_context_states:
                entity = self._mdib.entities.by_handle(proposed_st.DescriptorHandle)
                modified_entities.append(entity)
                old_state_container = None
                if proposed_st.DescriptorHandle != proposed_st.Handle:
                    # this is an update for an existing state or a new one
                    old_state_container = entity.states.get(proposed_st.Handle)
                    if old_state_container is None:
                        msg = f'handle {proposed_st.Handle} not found'
                        raise ValueError(msg)
                if old_state_container is None:
                    # this is a new context state
                    # create a new unique handle
                    proposed_st.Handle = uuid.uuid4().hex
                    if proposed_st.ContextAssociation == pm_types.ContextAssociation.ASSOCIATED:
                        # disassociate existing states
                        handles = self._mdib.xtra.disassociate_all(entity,
                                                                   unbinding_mdib_version = mgr.new_mdib_version)
                        operation_target_handles.extend(handles)
                        modified_state_handles[entity.handle].extend(handles)
                        # set version and time in new state
                        proposed_st.BindingMdibVersion = mgr.new_mdib_version
                        proposed_st.BindingStartTime = time.time()
                    self._logger.info('new %s, DescriptorHandle=%s Handle=%s',
                                      proposed_st.NODETYPE.localname, proposed_st.DescriptorHandle, proposed_st.Handle)
                    # add to entity, and keep handle for later
                    entity.states[proposed_st.Handle] = proposed_st
                else:
                    # this is an update to an existing patient
                    # use "regular" way to update via transaction manager
                    self._logger.info('update %s, handle=%s', proposed_st.NODETYPE.localname, proposed_st.Handle)
                    # handle changed ContextAssociation
                    if (old_state_container.ContextAssociation == pm_types.ContextAssociation.ASSOCIATED
                            and proposed_st.ContextAssociation != pm_types.ContextAssociation.ASSOCIATED):
                        proposed_st.UnbindingMdibVersion = mgr.new_mdib_version
                        proposed_st.BindingEndTime = time.time()
                    elif (old_state_container.ContextAssociation != pm_types.ContextAssociation.ASSOCIATED
                          and proposed_st.ContextAssociation == pm_types.ContextAssociation.ASSOCIATED):
                        proposed_st.BindingMdibVersion = mgr.new_mdib_version
                        proposed_st.BindingStartTime = time.time()
                        handles = self._mdib.xtra.disassociate_all(entity,
                                                                   unbinding_mdib_version = mgr.new_mdib_version,
                                                                   ignored_handle=old_state_container.Handle)
                        operation_target_handles.extend(handles)
                        modified_state_handles[entity.handle].extend(handles)
                    old_state_container.update_from_other_container(proposed_st, skipped_properties=[
                                                                                           'BindingMdibVersion',
                                                                                           'UnbindingMdibVersion',
                                                                                           'BindingStartTime',
                                                                                           'BindingEndTime',
                                                                                           'StateVersion'])
                modified_state_handles[entity.handle].append(proposed_st.Handle)
                operation_target_handles.append(proposed_st.Handle)

            # write changes back to mdib
            for entity in modified_entities:
                handles = modified_state_handles[entity.handle]
                mgr.write_entity(entity, handles)

            if len(operation_target_handles) == 1:
                return ExecuteResult(operation_target_handles[0],
                                     self._mdib.data_model.msg_types.InvocationState.FINISHED)
            # the operation manipulated more than one context state, but the operation can only return a single handle.
            # (that is a BICEPS shortcoming, the string return type only reflects that situation).
            return ExecuteResult(params.operation_instance.operation_target_handle,
                                 self._mdib.data_model.msg_types.InvocationState.FINISHED)


class EnsembleContextProvider(GenericContextProvider):
    """EnsembleContextProvider."""

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str | None = None):
        super().__init__(mdib,
                         op_target_descr_types=[mdib.data_model.pm_names.EnsembleContextDescriptor],
                         log_prefix=log_prefix)


class LocationContextProvider(GenericContextProvider):
    """LocationContextProvider."""

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str | None = None):
        super().__init__(mdib,
                         op_target_descr_types=[mdib.data_model.pm_names.LocationContextDescriptor],
                         log_prefix=log_prefix)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from sdc11073.pysoap.msgreader import ReceivedMessage

class RequestData:
    """This class holds all information about the processing of a http request together"""

    def __init__(self,
                 http_header: dict,
                 path: str,
                 peer_name: str,
                 request: Optional[bytes] = None,
                 message_data: Optional[ReceivedMessage] = None):
        self.http_header: dict = http_header
        self.path = path
        self.peer_name: str = peer_name  # for logging
        self.request: Optional[bytes] = request
        self.message_data: Optional[ReceivedMessage] = message_data
        self.consumed_path_elements = []
        if path.startswith('/'):
            path = path[1:]
        self.path_elements = path.split('/')

    def consume_current_path_element(self):
        if len(self.path_elements) == 0:
            return None
        self.consumed_path_elements.append(self.path_elements[0])
        self.path_elements = self.path_elements[1:]
        return self.consumed_path_elements[-1]

    @property
    def current_path_element(self):
        return self.path_elements[0] if len(self.path_elements) > 0 else None



"""The module implements the SdcProvider."""
from __future__ import annotations

import copy
import threading
import uuid
from typing import TYPE_CHECKING, Any, Protocol
from urllib.parse import SplitResult

import sdc11073.certloader
from sdc11073 import loghelper
from sdc11073 import observableproperties as properties
from sdc11073.dispatch import (
    DispatchKey,
    MessageConverterMiddleware,
    PathElementRegistry,
    RequestData,
    RequestDispatcher,
)
from sdc11073.exceptions import ApiUsageError
from sdc11073.httpserver import compression
from sdc11073.httpserver.httpserverimpl import HttpServerThreadBase
from sdc11073.namespaces import WSA_ANONYMOUS
from sdc11073.pysoap.soapclientpool import SoapClientPool
from sdc11073.xml_types import mex_types
from sdc11073.xml_types.addressing_types import EndpointReferenceType
from sdc11073.xml_types.dpws_types import HostServiceType, ThisDeviceType, ThisModelType
from sdc11073.xml_types.wsd_types import ProbeMatchesType, ProbeMatchType

from .periodicreports import PeriodicReportsHandler, PeriodicReportsNullHandler

if TYPE_CHECKING:
    from enum import Enum

    from sdc11073.location import SdcLocation
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.mdib.statecontainers import AbstractStateProtocol
    from sdc11073.mdib.transactionsprotocol import TransactionResultProtocol
    from sdc11073.provider.porttypes.localizationservice import LocalizationStorage
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.pysoap.soapenvelope import ReceivedSoapMessage
    from sdc11073.roles.protocols import ProductProtocol, WaveformProviderProtocol
    from sdc11073.xml_types.msg_types import AbstractSet
    from sdc11073.xml_types.pm_types import InstanceIdentifier
    from sdc11073.xml_types.wsd_types import ScopesType

    from .components import SdcProviderComponents
    from .operations import OperationDefinitionBase

class _PathElementDispatcher(PathElementRegistry):
    """Dispatch to one of the registered instances, based on path element.

    Implements RequestHandlerProtocol.
    """

    def register_instance(self, path_element: str | None, instance: RequestDispatcher):
        super().register_instance(path_element, instance)

    def on_post(self, request_data: RequestData) -> CreatedMessage:
        path_element = request_data.consume_current_path_element()
        dispatcher = self.get_instance(path_element)
        return dispatcher.on_post(request_data)

    def on_get(self, request_data: RequestData) -> str:
        dispatcher = self.get_instance(request_data.consume_current_path_element())
        return dispatcher.on_get(request_data)


class WsDiscoveryProtocol(Protocol):
    """WsDiscoveryProtocol is the interface that SdcProvider expects."""

    def publish_service(self, epr: str, types: list, scopes: ScopesType, x_addrs: list):  # noqa: D102
        ...

    def get_active_addresses(self) -> list:  # noqa: D102
        ...

    def clear_service(self, epr: str):  # noqa: D102
        ...


class SdcProvider:
    """SdcProvider is the host for sdc services, subscription manager etc."""

    DEFAULT_CONTEXTSTATES_IN_GETMDIB = True  # defines weather get_mdib and getMdStates contain context states or not.

    def __init__(self, # noqa: PLR0915, PLR0913
                 ws_discovery: WsDiscoveryProtocol,
                 this_model: ThisModelType,
                 this_device: ThisDeviceType,
                 device_mdib_container: ProviderMdibProtocol,
                 epr: str | uuid.UUID | None = None,
                 validate: bool = True,
                 ssl_context_container: sdc11073.certloader.SSLContextContainer | None = None,
                 max_subscription_duration: int = 15,
                 socket_timeout: int | float | None = None, # noqa: PYI041
                 log_prefix: str = '',
                 default_components: SdcProviderComponents | None = None,
                 specific_components: SdcProviderComponents | None = None,
                 chunk_size: int = 0,
                 alternative_hostname: str | None = None):
        """Construct an SdcProvider.

        :param ws_discovery: a WsDiscovers instance
        :param this_model: a ThisModelType instance
        :param this_device: a ThisDeviceType instance
        :param device_mdib_container: a ProviderMdibProtocol instance
        :param epr: something that serves as a unique identifier of this device for discovery.
                    If epr is a string, it must be usable as a path element in an url (no spaces, ...)
        :param validate: bool
        :param ssl_context_container: if not None, the contexts are used and an https url is used, otherwise http
        :param max_subscription_duration: max. possible duration of a subscription
        :param socket_timeout: timeout for tcp sockets that send notifications.
                               If None, it is set to max_subscription_duration * 1.2
        :param log_prefix: a string
        :param specific_components: a SdcProviderComponents instance
        :param chunk_size: if value > 0, messages are split into chunks of this size.
        :param alternative_hostname: if supplied this hostname is used in xaddr, default is to use numerical
                                     ipv4 address (can be used to use full qualified hostname)
        """
        self._wsdiscovery = ws_discovery
        self.model = this_model
        self.device = this_device
        self._mdib = device_mdib_container
        if epr is None:
            self._epr = uuid.uuid4()
        else:
            self._epr = epr
        self._validate = validate
        self._ssl_context_container = ssl_context_container
        self._max_subscription_duration = max_subscription_duration
        self._socket_timeout = socket_timeout or int(max_subscription_duration * 1.2)
        self._log_prefix = log_prefix
        if default_components is None:
            from .components import default_sdc_provider_components  # lazy import avoids cyclic import
            default_components = default_sdc_provider_components
        self._components = copy.deepcopy(default_components)
        if specific_components is not None:
            # merge specific stuff into _components
            self._components.merge(specific_components)
        self.chunk_size = chunk_size
        self._alternative_hostname = alternative_hostname
        self._mdib.log_prefix = log_prefix
        self._compression_methods = compression.CompressionHandler.available_encodings[:]
        self._logger = loghelper.get_logger_adapter('sdc.device', log_prefix)
        self._location = None
        self._http_server = None
        self._is_internal_http_server = False

        if self._ssl_context_container is not None:
            self._urlschema = 'https'
        else:
            self._urlschema = 'http'

        self.collect_rt_samples_period = 0.1  # in seconds
        self.contextstates_in_getmdib = self.DEFAULT_CONTEXTSTATES_IN_GETMDIB  # can be overridden per instance
        # look for schemas added by services and components spec
        additional_schema_specs = set(self._components.additional_schema_specs)
        for hosted_service in self._components.hosted_services.values():
            for port_type_impl in hosted_service:
                additional_schema_specs.update(port_type_impl.additional_namespaces)
        logger = loghelper.get_logger_adapter('sdc.device.msgreader', log_prefix)
        self.msg_reader = self._components.msg_reader_class(self._mdib.sdc_definitions,
                                                            list(additional_schema_specs),
                                                            logger,
                                                            validate=validate)

        logger = loghelper.get_logger_adapter('sdc.device.msgfactory', log_prefix)
        self.msg_factory = self._components.msg_factory_class(self._mdib.sdc_definitions,
                                                              list(additional_schema_specs),
                                                              logger=logger,
                                                              validate=validate)

        # host dispatcher provides data of the sdc device itself.
        self._host_dispatcher = RequestDispatcher()
        nsh = self._mdib.sdc_definitions.data_model.ns_helper
        self._host_dispatcher.register_post_handler(
            DispatchKey(f'{nsh.WXF.namespace}/Get', None),
            self._on_get_metadata)
        self._host_dispatcher.register_post_handler(
            DispatchKey(f'{nsh.WSD.namespace}/Probe', nsh.WSD.tag('Probe')),
            self._on_probe_request)
        epr_type = EndpointReferenceType()
        epr_type.Address = self.epr_urn
        self.dpws_host = HostServiceType()
        self.dpws_host.EndpointReference = epr_type
        self.dpws_host.Types = self._mdib.sdc_definitions.MedicalDeviceTypesFilter

        self._hosted_service_dispatcher = _PathElementDispatcher()
        self._hosted_service_dispatcher.register_instance(None, self._host_dispatcher)

        self._msg_converter = MessageConverterMiddleware(
            self.msg_reader, self.msg_factory, self._logger, self._hosted_service_dispatcher)

        self._transaction_id = 0  # central transaction number handling for all called operations.
        self._transaction_id_lock = threading.Lock()

        # these are initialized in _setup_components:
        self._subscriptions_managers = {}
        self._soap_client_pool = SoapClientPool(self._mk_soap_client, log_prefix)
        self._sco_operations_registries = {}  # key is mds handle ?
        self._service_factory = None
        self.product_lookup: dict[str, ProductProtocol] = {}  # one product per sco,  key is a sco handle
        self.hosted_services = None
        self._periodic_reports_handler = PeriodicReportsNullHandler()
        self.waveform_provider: WaveformProviderProtocol | None = None
        self._setup_components()
        self.base_urls = []  # will be set after httpserver is started
        properties.bind(device_mdib_container, transaction=self._send_episodic_reports)
        properties.bind(device_mdib_container, rt_updates=self._send_rt_notifications)

    def generate_transaction_id(self) -> int:
        """Return a new transaction id."""
        with self._transaction_id_lock:
            self._transaction_id += 1
            return self._transaction_id

    def _mk_soap_client(self, netloc: str, accepted_encodings: list[str]) -> Any:
        cls = self._components.soap_client_class
        return cls(netloc,
                   self._socket_timeout,
                   loghelper.get_logger_adapter('sdc.device.soap', self._log_prefix),
                   ssl_context=self._ssl_context_container.client_context if self._ssl_context_container else None,
                   sdc_definitions=self._mdib.sdc_definitions,
                   msg_reader=self.msg_reader,
                   supported_encodings=self._compression_methods,
                   request_encodings=accepted_encodings,
                   chunk_size=self.chunk_size)

    def _setup_components(self):
        self._subscriptions_managers = {}
        for name, cls in self._components.subscriptions_manager_class.items():
            mgr = cls(self._mdib.sdc_definitions,
                      self.msg_factory,
                      self._soap_client_pool,
                      self._max_subscription_duration,
                      log_prefix=self._log_prefix,
                      )
            self._subscriptions_managers[name] = mgr

        services_factory = self._components.services_factory
        self.hosted_services = services_factory(self, self._components, self._subscriptions_managers)
        for dpws_service in self.hosted_services.dpws_hosted_services.values():
            self._hosted_service_dispatcher.register_instance(dpws_service.path_element, dpws_service)

        cls = self._components.sco_operations_registry_class
        pm_names = self._mdib.data_model.pm_names

        entities = self._mdib.entities.by_node_type(pm_names.ScoDescriptor)
        for entity in entities:
            sco_operations_registry = cls(self.hosted_services.set_service,
                                          self._components.operation_cls_getter,
                                          self._mdib,
                                          entity.descriptor,
                                          log_prefix=self._log_prefix)
            self._sco_operations_registries[entity.handle] = sco_operations_registry

            product_roles = self._components.role_provider_class(self._mdib,
                                                                 sco_operations_registry,
                                                                 self._log_prefix)
            self.product_lookup[entity.handle] = product_roles
            product_roles.init_operations()
        if self._components.waveform_provider_class is not None:
            self.waveform_provider = self._components.waveform_provider_class(self._mdib,
                                                                              self._log_prefix)

        # product roles might have added descriptors, set source mds for all
        self._mdib.xtra.set_all_source_mds()

    @property
    def localization_storage(self) -> LocalizationStorage | None:
        """Convenience method for easier access to LocalizationStorage."""
        if self.hosted_services.localization_service is not None:
            return self.hosted_services.localization_service.localization_storage
        return None

    def _on_get_metadata(self, request_data: RequestData) -> CreatedMessage:
        self._logger.info('_on_get_metadata from %s', request_data.peer_name)
        metadata = mex_types.Metadata()
        section = mex_types.ThisModelMetadataSection()
        section.MetadataReference = self.model
        metadata.MetadataSection.append(section)

        section = mex_types.ThisDeviceMetadataSection()
        section.MetadataReference = self.device
        metadata.MetadataSection.append(section)

        section = mex_types.RelationshipMetadataSection()
        section.MetadataReference.Host = self.dpws_host

        # add all hosted services:
        for service in self.hosted_services.dpws_hosted_services.values():
            hosted = service.mk_dpws_hosted_instance()
            section.MetadataReference.Hosted.append(hosted)
        metadata.MetadataSection.append(section)

        # find namespaces that are used in Types of Host and Hosted
        _nsm = self._mdib.nsmapper
        needed_namespaces = [_nsm.DPWS, _nsm.WSX]
        q_names = []
        q_names.extend(self.dpws_host.Types)
        for h in section.MetadataReference.Hosted:
            q_names.extend(h.Types)
        for q_name in q_names:
            for e in _nsm.prefix_enum:
                if e.namespace == q_name.namespace and e not in needed_namespaces:
                    needed_namespaces.append(e)
        return self.msg_factory.mk_reply_soap_message(request_data, metadata, needed_namespaces)

    def _on_probe_request(self, request: RequestData) -> CreatedMessage:
        _nsm = self._mdib.nsmapper
        probe_matches = ProbeMatchesType()
        probe_match = ProbeMatchType()
        probe_match.Types.append(_nsm.DPWS.tag('Device'))
        probe_match.Types.append(_nsm.MDPWS.tag('MedicalDevice'))
        probe_match.XAddrs.extend(self.get_xaddrs())
        probe_matches.ProbeMatch.append(probe_match)
        needed_namespaces = [_nsm.DPWS, _nsm.MDPWS]
        response = self.msg_factory.mk_reply_soap_message(request, probe_matches, needed_namespaces)
        response.p_msg.header_info_block.To = WSA_ANONYMOUS
        return response

    def set_location(self,
                     location: SdcLocation,
                     validators: list[InstanceIdentifier] | None = None,
                     publish_now: bool = True,
                     location_context_descriptor_handle: str | None = None):
        """Set a new associated location.

        :param location: an SdcLocation instance
        :param validators: a list of InstanceIdentifier objects or None;
            If it is None, the defaultInstanceIdentifiers member is used
        :param publish_now: if True, the device is published via its wsdiscovery reference.
        :param location_context_descriptor_handle: Only needed if the mdib contains more than one
               LocationContextDescriptor. Then this defines the descriptor for which a new LocationContextState
               shall be created.

        """
        if location == self._location:
            return
        self._location = location
        self._mdib.xtra.set_location(location,
                                     validators,
                                     location_context_descriptor_handle = location_context_descriptor_handle)
        if publish_now:
            self.publish()

    def publish(self):
        """Publish device on the network (sends HELLO message)."""
        scopes = self._components.scopes_factory(self._mdib)
        x_addrs = self.get_xaddrs()
        self._wsdiscovery.publish_service(self.epr_urn,
                                          list(self._mdib.sdc_definitions.MedicalDeviceTypesFilter),
                                          scopes,
                                          x_addrs)

    @property
    def mdib(self) -> ProviderMdibProtocol:
        """Return mdib reference."""
        return self._mdib

    @property
    def epr_urn(self) -> str:
        """Return end point reference, e.g 'urn:uuid:8c26f673-fdbf-4380-b5ad-9e2454a65b6b'."""
        try:
            return self._epr.urn
        except AttributeError:
            return self._epr

    @property
    def path_prefix(self) -> str:
        """Return http path prefix of service e.g '8c26f673fdbf4380b5ad9e2454a65b6b'."""
        try:
            return self._epr.hex
        except AttributeError:
            return self._epr

    def get_operation_by_handle(self, operation_handle: str) -> OperationDefinitionBase | None:
        """Return OperationDefinitionBase for given handle or None if it does not exist."""
        for sco in self._sco_operations_registries.values():
            op = sco.get_operation_by_handle(operation_handle)
            if op is not None:
                return op
        return None

    def handle_operation_request(self,
                                 operation: OperationDefinitionBase,
                                 request: ReceivedSoapMessage,
                                 operation_request: AbstractSet,
                                 transaction_id: int) -> Enum:
        """Find the responsible sco and forward request to it."""
        for sco in self._sco_operations_registries.values():
            has_this_operation = sco.get_operation_by_handle(operation.handle) is not None
            if has_this_operation:
                return sco.handle_operation_request(operation, request, operation_request, transaction_id)
        self._logger.error('no sco has operation %s', operation.handle)
        return self.mdib.data_model.msg_types.InvocationState.FAILED

    def start_all(self,
                  start_rtsample_loop: bool = True,
                  periodic_reports_interval: float | None = None,
                  shared_http_server=None): # noqa: ANN001
        """Start all background threads.

        :param start_rtsample_loop: flag
        :param periodic_reports_interval: if provided, a value in seconds
        :param shared_http_server: if provided, use this http server, else device creates its own.
        :return:
        """
        if periodic_reports_interval or self._mdib.retrievability_periodic:
            self._logger.info('starting PeriodicReportsHandler')
            self._periodic_reports_handler = PeriodicReportsHandler(self._mdib,
                                                                    self.hosted_services,
                                                                    periodic_reports_interval)
            self._periodic_reports_handler.start()
        else:
            self._logger.info('no PeriodicReportsHandler')
            self._periodic_reports_handler = PeriodicReportsNullHandler()
        self._start_services(shared_http_server)

        if start_rtsample_loop:
            self.start_rt_sample_loop()

    def _start_services(self, shared_http_server=None): # noqa: ANN001
        """Start the services."""
        self._logger.info('starting services, addr = %r', self._wsdiscovery.get_active_addresses())
        for sco in self._sco_operations_registries.values():
            sco.start_worker()

        if shared_http_server:
            self._http_server = shared_http_server
        else:
            self._is_internal_http_server = True
            logger = loghelper.get_logger_adapter('sdc.device.httpsrv', self._log_prefix)

            self._http_server = HttpServerThreadBase(
                my_ipaddress='0.0.0.0',  # noqa: S104
                ssl_context=self._ssl_context_container.server_context if self._ssl_context_container else None,
                supported_encodings=self._compression_methods,
                logger=logger,
                chunk_size=self.chunk_size)

            # first start http server, the services need to know the ip port number
            self._http_server.start()
            event_is_set = self._http_server.started_evt.wait(timeout=15.0)
            if not event_is_set:
                self._logger.error('Cannot start device, start event of http server not set.')
                raise RuntimeError('Cannot start device, start event of http server not set.')

        host_ips = self._wsdiscovery.get_active_addresses()
        self._http_server.dispatcher.register_instance(self.path_prefix, self._msg_converter)
        if len(host_ips) == 0:
            self._logger.error('Cannot start device, there is no IP address to bind it to.')
            raise RuntimeError('Cannot start device, there is no IP address to bind it to.')

        port = self._http_server.my_port
        if port is None:
            self._logger.error('Cannot start device, could not bind HTTP server to a port.')
            raise RuntimeError('Cannot start device, could not bind HTTP server to a port.')

        self.base_urls = []
        for addr in host_ips:
            self.base_urls.append(
                SplitResult(self._urlschema, f'{addr}:{port}', self.path_prefix, query=None, fragment=None))

        for host_ip in host_ips:
            self._logger.info('serving Services on %s:%d', host_ip, port)
        for subscriptions_manager in self._subscriptions_managers.values():
            subscriptions_manager.set_base_urls(self.base_urls)

    def stop_all(self, send_subscription_end: bool = True):
        """Stop all background threads and clear local data."""
        self.stop_realtime_sample_loop()
        if self._periodic_reports_handler:
            self._periodic_reports_handler.stop()
        for subscriptions_manager in self._subscriptions_managers.values():
            subscriptions_manager.stop_all(send_subscription_end)
        for sco in self._sco_operations_registries.values():
            sco.stop_worker()
        try:
            self._wsdiscovery.clear_service(self.epr_urn)
        except KeyError:
            self._logger.info('epr "%s" not known in self._wsdiscovery', self.epr_urn)
        for role in self.product_lookup.values():
            role.stop()
        if self._is_internal_http_server and self._http_server is not None:
            self._http_server.stop()
        self._soap_client_pool.close_all()

    def start_rt_sample_loop(self):
        """Start generating waveform data."""
        if self.waveform_provider is None:
            raise ApiUsageError('no waveform provider configured.')
        if self.waveform_provider.is_running:
            raise ApiUsageError('realtime send loop already started')
        self.waveform_provider.start()

    def stop_realtime_sample_loop(self):
        """Stop generating waveform data."""
        if self.waveform_provider is not None and self.waveform_provider.is_running:
            self.waveform_provider.stop()

    def get_xaddrs(self) -> list[str]:
        """Return the addresses of the provider."""
        if self._alternative_hostname:
            addresses = [self._alternative_hostname]
        else:
            # these own IP addresses are currently used by discovery
            addresses = self._wsdiscovery.get_active_addresses()

        port = self._http_server.my_port
        xaddrs = []
        for addr in addresses:
            xaddrs.append(f'{self._urlschema}://{addr}:{port}/{self.path_prefix}') # noqa: PERF401
        return xaddrs

    def _send_episodic_reports(self, transaction_result: TransactionResultProtocol):
        mdib_version_group = self._mdib.mdib_version_group
        if transaction_result.has_descriptor_updates:
            port_type_impl = self.hosted_services.description_event_service
            updated = transaction_result.descr_updated
            created = transaction_result.descr_created
            deleted = transaction_result.descr_deleted
            states = transaction_result.all_states()
            port_type_impl.send_descriptor_updates(
                updated, created, deleted, states, mdib_version_group)

        states = transaction_result.metric_updates
        if len(states) > 0:
            port_type_impl = self.hosted_services.state_event_service
            port_type_impl.send_episodic_metric_report(
                states, mdib_version_group)
            self._periodic_reports_handler.store_metric_states(mdib_version_group.mdib_version,
                                                               transaction_result.metric_updates)

        states = transaction_result.alert_updates
        if len(states) > 0:
            port_type_impl = self.hosted_services.state_event_service
            port_type_impl.send_episodic_alert_report(
                states, mdib_version_group)
            self._periodic_reports_handler.store_alert_states(mdib_version_group.mdib_version, states)

        states = transaction_result.comp_updates
        if len(states) > 0:
            port_type_impl = self.hosted_services.state_event_service
            port_type_impl.send_episodic_component_state_report(
                states, mdib_version_group)
            self._periodic_reports_handler.store_component_states(mdib_version_group.mdib_version, states)

        states = transaction_result.ctxt_updates
        if len(states) > 0:
            port_type_impl = self.hosted_services.context_service
            port_type_impl.send_episodic_context_report(
                states, mdib_version_group)
            self._periodic_reports_handler.store_context_states(mdib_version_group.mdib_version, states)

        states = transaction_result.op_updates
        if len(states) > 0:
            port_type_impl = self.hosted_services.state_event_service
            port_type_impl.send_episodic_operational_state_report(states, mdib_version_group)
            self._periodic_reports_handler.store_operational_states(mdib_version_group.mdib_version, states)

        states = transaction_result.rt_updates
        if len(states) > 0:
            port_type_impl = self.hosted_services.waveform_service
            port_type_impl.send_realtime_samples_report(states, mdib_version_group)

    def _send_rt_notifications(self, rt_states: list[AbstractStateProtocol]):
        if len(rt_states) > 0:
            port_type_impl = self.hosted_services.waveform_service
            port_type_impl.send_realtime_samples_report(rt_states, self._mdib.mdib_version_group)

    def set_used_compression(self, *compression_methods: str):
        """Set supported compression methods, e.g. 'gzip'."""
        del self._compression_methods[:]
        self._compression_methods.extend(compression_methods)



"""Implementation of clock provider functionality."""
from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from sdc11073.provider.operations import ExecuteResult

from .nomenclature import NomenclatureCodes
from .providerbase import OperationClassGetter, ProviderRole

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol, AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase
    from sdc11073.provider.sco import AbstractScoOperationsRegistry
    from sdc11073.xml_types.pm_types import CodedValue, SafetyClassification


class GenericSDCClockProvider(ProviderRole):
    """Handles operations for setting ntp server and time zone.

    This provider handles SetString operations with codes
    "MDC_OP_SET_TIME_SYNC_REF_SRC" and "MDC_ACT_SET_TIME_ZONE".
    Nothing is added to the mdib. If the mdib does not contain these operations, the functionality is not available.
    """

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        super().__init__(mdib, log_prefix)
        self._set_ntp_operations = []
        self._set_tz_operations = []
        pm_types = self._mdib.data_model.pm_types

        self.MDC_OP_SET_TIME_SYNC_REF_SRC = pm_types.CodedValue(NomenclatureCodes.MDC_OP_SET_TIME_SYNC_REF_SRC)
        self.MDC_ACT_SET_TIME_ZONE = pm_types.CodedValue(NomenclatureCodes.MDC_ACT_SET_TIME_ZONE)

    def init_operations(self, sco: AbstractScoOperationsRegistry):
        """Create a ClockDescriptor and ClockState in mdib if they do not exist in mdib."""
        super().init_operations(sco)
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names
        clock_entities = self._mdib.entities.by_node_type(pm_names.ClockDescriptor)
        if len(clock_entities) == 0:
            mds_entities = self._mdib.entitiesby_.node_type(pm_names.MdsDescriptor)
            if len(mds_entities) == 0:
                self._logger.info('empty mdib, cannot create a clock descriptor')
                return
            # create a clock descriptor for the first mds
            my_mds_entity = mds_entities[0]
            clock_descr_handle = 'clock_' + uuid.uuid4().hex
            self._logger.debug('creating a clock descriptor, handle=%s', clock_descr_handle)
            model = self._mdib.data_model
            clock_entity = self._mdib.entities.new_entity(model.pm_names.ClockDescriptor,
                                                          handle = clock_descr_handle,
                                                          parent_handle=my_mds_entity.handle)
            clock_entity.descriptor.SafetyClassification = pm_types.SafetyClassification.INF
            clock_entity.descriptor.Type = pm_types.CodedValue('123')
            with self._mdib.descriptor_transaction() as mgr:
                mgr.write_entity(clock_entity)

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Create operation handlers.

        Handle codes MDC_OP_SET_TIME_SYNC_REF_SRC, MDC_ACT_SET_TIME_ZONE.
        """
        if operation_descriptor_container.coding == self.MDC_OP_SET_TIME_SYNC_REF_SRC.coding:
            self._logger.debug('instantiating "set ntp server" operation from existing descriptor handle=%s',
                               operation_descriptor_container.Handle)
            set_ntp_operation = self._mk_operation_from_operation_descriptor(operation_descriptor_container,
                                                                             operation_cls_getter,
                                                                             operation_handler=self._set_ntp_string)
            self._set_ntp_operations.append(set_ntp_operation)
            return set_ntp_operation
        if operation_descriptor_container.coding == self.MDC_ACT_SET_TIME_ZONE.coding:
            self._logger.debug('instantiating "set time zone" operation from existing descriptor handle=%s',
                               operation_descriptor_container.Handle)
            set_tz_operation = self._mk_operation_from_operation_descriptor(operation_descriptor_container,
                                                                            operation_cls_getter,
                                                                            operation_handler=self._set_tz_string)
            self._set_tz_operations.append(set_tz_operation)
            return set_tz_operation
        return None

    def _set_ntp_string(self, params: ExecuteParameters) -> ExecuteResult:
        """Set the ReferenceSource value of clock state (ExecuteHandler)."""
        value = params.operation_request.argument
        pm_names = self._mdib.data_model.pm_names
        self._logger.info('set value %s from %s to %s',
                          params.operation_instance.operation_target_handle,
                          params.operation_instance.current_value, value)

        op_target_entity = self._mdib.entities.by_handle(params.operation_instance.operation_target_handle)

        # look for clock entities that are a direct child of this mds
        mds_handle = op_target_entity.descriptor.source_mds or op_target_entity.handle
        clock_entities = self._mdib.entities.by_node_type(pm_names.ClockDescriptor)
        clock_entities = [c for c in clock_entities if c.parent_handle == mds_handle]

        if len(clock_entities) == 0:
            self._logger.warning('_set_ntp_string: no clock entity found')
            return ExecuteResult(params.operation_instance.operation_target_handle,
                                 self._mdib.data_model.msg_types.InvocationState.FAILED,
                                 )

        clock_entities[0].state.ReferenceSource = [value]
        with self._mdib.component_state_transaction() as mgr:
            mgr.write_entity(clock_entities[0])
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def _set_tz_string(self, params: ExecuteParameters) -> ExecuteResult:
        """Set the TimeZone value of clock state (ExecuteHandler)."""
        value = params.operation_request.argument
        pm_names = self._mdib.data_model.pm_names
        self._logger.info('set value %s from %s to %s',
                          params.operation_instance.operation_target_handle,
                          params.operation_instance.current_value, value)

        op_target_entity = self._mdib.entities.by_handle(params.operation_instance.operation_target_handle)

        # look for clock entities that are a direct child of this mds
        mds_handle = op_target_entity.descriptor.source_mds or op_target_entity.handle
        clock_entities = self._mdib.entities.by_node_type(pm_names.ClockDescriptor)
        clock_entities = [c for c in clock_entities if c.parent_handle == mds_handle]

        if len(clock_entities) == 0:
            self._logger.warning('_set_ntp_string: no clock entity found')
            return ExecuteResult(params.operation_instance.operation_target_handle,
                                 self._mdib.data_model.msg_types.InvocationState.FAILED)

        clock_entities[0].state.TimeZone = value
        with self._mdib.component_state_transaction() as mgr:
            mgr.write_entity(clock_entities[0])
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def _create_clock_descriptor_container(self, handle: str,
                                           parent_handle: str,
                                           coded_value: CodedValue,
                                           safety_classification: SafetyClassification) -> AbstractDescriptorProtocol:
        """Create a ClockDescriptorContainer with the given properties.

        :param handle: Handle of the new container
        :param parent_handle: Handle of the parent
        :param coded_value: a pmtypes.CodedValue instance that defines what this onject represents in medical terms.
        :param safety_classification: a pmtypes.SafetyClassification value
        :return: the created object
        """
        model = self._mdib.data_model
        cls = model.get_descriptor_container_class(model.pm_names.ClockDescriptor)
        return self._create_descriptor_container(cls,
                                                 handle,
                                                 parent_handle,
                                                 coded_value,
                                                 safety_classification)


class SDCClockProvider(GenericSDCClockProvider):
    """SDCClockProvider adds SetString operations to set ntp server and time zone if they do not exist.

    This provider guarantees that there are SetString operations with codes "MDC_OP_SET_TIME_SYNC_REF_SRC"
    and "MDC_ACT_SET_TIME_ZONE" if mdib contains a ClockDescriptor. It adds them to mdib if they do not exist.

    """

    def make_missing_operations(self, sco: AbstractScoOperationsRegistry) -> list[OperationDefinitionBase]:
        """Add operations to mdib if mdib contains a ClockDescriptor, but not the operations."""
        pm_names = self._mdib.data_model.pm_names
        ops = []
        operation_cls_getter = sco.operation_cls_getter

        mds_container = self._mdib.descriptions.NODETYPE.get_one(pm_names.MdsDescriptor)
        clock_descriptor = self._mdib.descriptions.NODETYPE.get_one(pm_names.ClockDescriptor,
                                                                    allow_none=True)
        if clock_descriptor is None:
            # there is no clock element in mdib,
            return ops
        set_string_op_cls = operation_cls_getter(pm_names.SetStringOperationDescriptor)

        if not self._set_ntp_operations:
            self._logger.debug('adding "set ntp server" operation, code = %r',
                               NomenclatureCodes.MDC_OP_SET_TIME_SYNC_REF_SRC)
            set_ntp_operation = set_string_op_cls('SET_NTP_SRV_' + mds_container.handle,
                                                  clock_descriptor.handle,
                                                  self._set_ntp_string,
                                                  coded_value=self.MDC_OP_SET_TIME_SYNC_REF_SRC)
            self._set_ntp_operations.append(set_ntp_operation)
            ops.append(set_ntp_operation)
        if not self._set_tz_operations:
            self._logger.debug('adding "set time zone" operation, code = %r',
                               NomenclatureCodes.MDC_ACT_SET_TIME_ZONE)
            set_tz_operation = set_string_op_cls('SET_TZONE_' + mds_container.handle,
                                                 clock_descriptor.handle,
                                                 self._set_tz_string,
                                                 coded_value=self.MDC_ACT_SET_TIME_ZONE)
            self._set_tz_operations.append(set_tz_operation)
            ops.append(set_tz_operation)
        return ops



"""The module contains the implementations of transactions for ProviderMdib."""
from __future__ import annotations

import copy
import time
import uuid
from typing import TYPE_CHECKING, cast

from sdc11073.exceptions import ApiUsageError

from .statecontainers import AbstractMultiStateProtocol
from .transactionsprotocol import (
    AnyTransactionManagerProtocol,
    TransactionItem,
    TransactionResultProtocol,
    TransactionType,
)

if TYPE_CHECKING:
    from sdc11073.loghelper import LoggerAdapter

    from .descriptorcontainers import AbstractDescriptorProtocol
    from .mdibbase import Entity, MultiStateEntity
    from .providermdib import ProviderMdib
    from .statecontainers import AbstractStateProtocol

class _TransactionBase:
    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        self._mdib = device_mdib_container
        # provide the new mdib version that the commit of this transaction will create
        self.new_mdib_version = device_mdib_container.mdib_version + 1
        self._logger = logger
        self.descriptor_updates: dict[str, TransactionItem] = {}
        self.metric_state_updates: dict[str, TransactionItem] = {}
        self.alert_state_updates: dict[str, TransactionItem] = {}
        self.component_state_updates: dict[str, TransactionItem] = {}
        self.context_state_updates: dict[str, TransactionItem] = {}
        self.operational_state_updates: dict[str, TransactionItem] = {}
        self.rt_sample_state_updates: dict[str, TransactionItem] = {}
        self._error = False

    def _handle_state_updates(self, state_updates_dict: dict) -> list[TransactionItem]:
        """Update mdib table and return a list of states to be sent in notifications."""
        updates_list = []
        for transaction_item in state_updates_dict.values():
            if transaction_item.old is not None:
                table = self._mdib.context_states if transaction_item.old.is_context_state else self._mdib.states
                table.remove_object_no_lock(transaction_item.old)
            else:
                table = self._mdib.context_states if transaction_item.new.is_context_state else self._mdib.states
            table.add_object_no_lock(transaction_item.new)
            updates_list.append(transaction_item.new.mk_copy(copy_node=False))
        return updates_list

    def get_state_transaction_item(self, handle: str) -> TransactionItem | None:
        """If transaction has a state with given handle, return the transaction-item, otherwise None.

        :param handle: the Handle of a context state or the DescriptorHandle in all other cases
        """
        if not handle:
            raise ValueError('No handle for state specified')
        for lookup in (self.metric_state_updates,
                       self.alert_state_updates,
                       self.component_state_updates,
                       self.context_state_updates,
                       self.operational_state_updates,
                       self.rt_sample_state_updates):
            if handle in lookup:
                return lookup[handle]
        return None

    @property
    def error(self) -> bool:
        return self._error


class DescriptorTransaction(_TransactionBase):
    """A Transaction that allows to insert / update / delete Descriptors and to modify states related to them."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)

    def actual_descriptor(self, descriptor_handle: str) -> AbstractDescriptorProtocol:
        """Return the actual descriptor in open transaction or from mdib.

        This method does not add the descriptor to the transaction!
        The descriptor can already be part of the transaction, and e.g. in pre_commit handlers of role providers
        it can be necessary to have access to it.
        """
        if not descriptor_handle:
            raise ValueError('No handle for descriptor specified')
        tr_container = self.descriptor_updates.get(descriptor_handle)
        if tr_container is not None:
            if tr_container.new is None:  # descriptor is deleted in this transaction!
                msg = f'The descriptor {descriptor_handle} is going to be deleted'
                raise ValueError(msg)
            return tr_container.new
        return self._mdib.descriptions.handle.get_one(descriptor_handle)

    def add_descriptor(self,
                       descriptor_container: AbstractDescriptorProtocol,
                       adjust_descriptor_version: bool = True,
                       state_container: AbstractStateProtocol | None = None):
        """Add a new descriptor to mdib."""
        descriptor_handle = descriptor_container.Handle
        if descriptor_handle in self.descriptor_updates:
            msg = f'Descriptor {descriptor_handle} already in updated set!'
            raise ValueError(msg)
        if descriptor_handle in self._mdib.descriptions.handle:
            msg = f'Cannot create descriptor {descriptor_handle}, it already exists in mdib!'
            raise ValueError(msg)
        if adjust_descriptor_version:
            self._mdib.descriptions.set_version(descriptor_container)
        if descriptor_container.source_mds is None:
            self._mdib.xtra.set_source_mds(descriptor_container)
        self.descriptor_updates[descriptor_handle] = TransactionItem(None, descriptor_container)
        if state_container is not None:
            if state_container.DescriptorHandle != descriptor_handle:
                msg = f'State {state_container.DescriptorHandle} does not match descriptor {descriptor_handle}!'
                raise ValueError(msg)
            self.add_state(state_container)

    def remove_descriptor(self, descriptor_handle: str):
        """Remove existing descriptor from mdib."""
        if not descriptor_handle:
            raise ValueError('No handle for descriptor specified')
        if descriptor_handle in self.descriptor_updates:
            msg = f'Descriptor {descriptor_handle} already in updated set!'
            raise ValueError(msg)
        orig_descriptor_container = self._mdib.descriptions.handle.get_one(descriptor_handle)
        self.descriptor_updates[descriptor_handle] = TransactionItem(orig_descriptor_container, None)

    def get_descriptor(self, descriptor_handle: str) -> AbstractDescriptorProtocol:
        """Get a descriptor from mdib."""
        if not descriptor_handle:
            raise ValueError('No handle for descriptor specified')
        if descriptor_handle in self.descriptor_updates:
            msg = f'Descriptor {descriptor_handle} already in updated set!'
            raise ValueError(msg)
        orig_descriptor_container = self._mdib.descriptions.handle.get_one(descriptor_handle)
        descriptor_container = orig_descriptor_container.mk_copy()
        descriptor_container.increment_descriptor_version()
        self.descriptor_updates[descriptor_handle] = TransactionItem(orig_descriptor_container, descriptor_container)
        return descriptor_container

    def get_state(self, descriptor_handle: str) -> AbstractStateProtocol:
        """Read a state from mdib and add it to the transaction.

        This method only allows to get a state if the corresponding descriptor is already part of the transaction.
        if not, it raises an ApiUsageError.
        """
        if not descriptor_handle:
            raise ValueError('No handle for state specified')
        if descriptor_handle not in self.descriptor_updates:
            raise ApiUsageError('Transaction does not contain the corresponding descriptor!')
        descriptor = self.descriptor_updates[descriptor_handle].new
        if descriptor.is_context_descriptor:
            # prevent this for simplicity reasons
            raise ApiUsageError('Transaction does not support extra handling of context states!')

        updates_dict = self._get_states_update(descriptor)
        if descriptor_handle in updates_dict:
            msg = f'State {descriptor_handle} already in updated set!'
            raise ValueError(msg)

        mdib_state = self._mdib.states.descriptor_handle.get_one(descriptor_handle, allow_none=False)
        copied_state = mdib_state.mk_copy()
        copied_state.increment_state_version()
        updates_dict[descriptor_handle] = TransactionItem(mdib_state, copied_state)
        return copied_state

    def add_state(self, state_container: AbstractStateProtocol, adjust_state_version: bool = True):
        """Add a new state to mdib.

        This method only allows to add a state if the corresponding descriptor is already part of the transaction.
        if not, it raises an ApiUsageError.
        """
        if state_container.DescriptorHandle not in self.descriptor_updates:
            raise ApiUsageError('Transaction has no descriptor for this state!')
        updates_dict = self._get_states_update(state_container)

        if state_container.is_context_state:
            if state_container.Handle is None:
                state_container.Handle = uuid.uuid4().hex
            key = state_container.Handle
        else:
            key = state_container.DescriptorHandle

        if key in updates_dict:
            msg = f'State {key} already in updated set!'
            raise ValueError(msg)

        # set reference to descriptor
        state_container.descriptor_container = self.descriptor_updates[state_container.DescriptorHandle].new
        state_container.DescriptorVersion = state_container.descriptor_container.DescriptorVersion
        if adjust_state_version:
            if state_container.is_context_state:
                self._mdib.context_states.set_version(state_container)
            else:
                self._mdib.states.set_version(state_container)
        updates_dict[key] = TransactionItem(None, state_container)

    def write_entity(self, # noqa: PLR0912, C901
                     entity: Entity | MultiStateEntity,
                     adjust_version_counter: bool = True):
        """Insert or update an entity."""
        descriptor_handle = entity.descriptor.Handle
        if descriptor_handle in self.descriptor_updates:
            msg = f'Entity {descriptor_handle} already in updated set!'
            raise ValueError(msg)

        tmp_descriptor = copy.deepcopy(entity.descriptor)
        orig_descriptor_container = self._mdib.descriptions.handle.get_one(descriptor_handle, allow_none=True)

        if adjust_version_counter:
            if orig_descriptor_container is None:
                # new descriptor, update version from saved versions in mdib if exists
                self._mdib.descriptions.set_version(tmp_descriptor)
            else:
                # update from old
                tmp_descriptor.DescriptorVersion = orig_descriptor_container.DescriptorVersion + 1

        self.descriptor_updates[descriptor_handle] = TransactionItem(orig_descriptor_container,
                                                                     tmp_descriptor)

        if entity.is_multi_state:
            old_states = self._mdib.context_states.descriptor_handle.get(descriptor_handle, [])
            old_states_dict = {s.Handle: s for s in old_states}
            for state_container in entity.states.values():
                tmp_state = copy.deepcopy(state_container)
                old_state = old_states_dict.get(tmp_state.Handle) # can be None => new state
                if adjust_version_counter:
                    tmp_state.DescriptorVersion = tmp_descriptor.DescriptorVersion
                    if old_state is not None:
                        tmp_state.StateVersion = old_state.StateVersion + 1
                    else:
                        self._mdib.context_states.set_version(tmp_state)

                self.context_state_updates[state_container.Handle] = TransactionItem(old_state, tmp_state)
            deleted_states_handles = set(old_states_dict.keys()).difference(set(entity.states.keys()))
            for handle in deleted_states_handles:
                del_state = old_states_dict[handle]
                self.context_state_updates[handle] = TransactionItem(del_state, None)
        else:
            tmp_state = copy.deepcopy(entity.state)
            tmp_state.descriptor_container = tmp_descriptor
            old_state = self._mdib.states.descriptor_handle.get_one(descriptor_handle, allow_none=True)
            if adjust_version_counter:
                if old_state is not None:
                    tmp_state.StateVersion = old_state.StateVersion + 1
                else:
                    self._mdib.states.set_version(tmp_state)

            state_updates_dict = self._get_states_update(tmp_state)
            state_updates_dict[entity.state.DescriptorHandle] = TransactionItem(old_state, tmp_state)

    def write_entities(self, entities: list[Entity | MultiStateEntity], adjust_version_counter: bool = True):
        """Write entities in order parents first."""
        written_handles = []
        ent_dict = {ent.handle: ent for ent in entities}
        while len(written_handles) < len(ent_dict):
            for handle, ent in ent_dict.items():
                write_now = not(ent.parent_handle is not None
                                and ent.parent_handle in ent_dict
                                and ent.parent_handle not in written_handles)
                if write_now and handle not in written_handles:
                    # it has a parent, and parent has not been written yet
                    self.write_entity(ent, adjust_version_counter)
                    written_handles.append(handle)

    def remove_entity(self, entity: Entity | MultiStateEntity):
        """Remove existing entity from mdib."""
        self.remove_descriptor(entity.handle)

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult.

        The parameter set_determination_time is only present in order to implement the interface correctly.
        Determination time is not set, because descriptors have no modification time.
        """
        proc = TransactionResult()
        if self.descriptor_updates:
            self._mdib.mdib_version = self.new_mdib_version
            # need to know all to be deleted and to be created descriptors
            to_be_deleted_handles = [tr_item.old.Handle for tr_item in self.descriptor_updates.values()
                                     if tr_item.new is None and tr_item.old is not None]
            to_be_created_handles = [tr_item.new.Handle for tr_item in self.descriptor_updates.values()
                                     if tr_item.old is None and tr_item.new is not None]
            # Remark 1:
            # handling only updated states here: If a descriptor is created, it can be assumed that the
            # application also creates the state in a transaction.
            # The state will then be transported via that notification report.
            # Maybe this needs to be reworked, but at the time of this writing it seems fine.
            #
            # Remark 2:
            # DescriptionModificationReport also contains the states that are related to the descriptors.
            # => if there is one, update its DescriptorVersion and add it to list of states that shall be sent
            # (Assuming that context descriptors (patient, location) are never changed,
            #  additional check for states in self.context_states is not needed.
            #  If this assumption is wrong, that functionality must be added!)

            for tr_item in self.descriptor_updates.values():
                orig_descriptor, new_descriptor = tr_item.old, tr_item.new
                if orig_descriptor is None:
                    # this is a create operation
                    self._logger.debug(  # noqa: PLE1205
                        'transaction_manager: new descriptor Handle={}, DescriptorVersion={}',
                        new_descriptor.Handle, new_descriptor.DescriptorVersion)
                    proc.descr_created.append(new_descriptor.mk_copy())
                    self._mdib.descriptions.add_object_no_lock(new_descriptor)
                    # increment DescriptorVersion if a child descriptor is added or deleted.
                    if new_descriptor.parent_handle is not None \
                            and new_descriptor.parent_handle not in to_be_created_handles:
                        # only update parent if it is not also created in this transaction
                        self._increment_parent_descriptor_version(proc, new_descriptor)
                    self._update_corresponding_state(new_descriptor)
                elif new_descriptor is None:
                    # this is a delete operation
                    self._logger.debug(  # noqa: PLE1205
                        'transaction_manager: rm descriptor Handle={}, DescriptorVersion={}',
                        orig_descriptor.Handle, orig_descriptor.DescriptorVersion)
                    all_descriptors = self._mdib.get_all_descriptors_in_subtree(orig_descriptor)
                    self._mdib.rm_descriptors_and_states(all_descriptors)
                    proc.descr_deleted.extend([d.mk_copy() for d in all_descriptors])
                    # increment DescriptorVersion if a child descriptor is added or deleted.
                    if orig_descriptor.parent_handle is not None \
                            and orig_descriptor.parent_handle not in to_be_deleted_handles:
                        # only update parent if it is not also deleted in this transaction
                        self._increment_parent_descriptor_version(proc, orig_descriptor)
                else:
                    # this is an update operation
                    proc.descr_updated.append(new_descriptor)
                    self._logger.debug(  # noqa: PLE1205
                        'transaction_manager: update descriptor Handle={}, DescriptorVersion={}',
                        new_descriptor.Handle, new_descriptor.DescriptorVersion)
                    orig_descriptor.update_from_other_container(new_descriptor)
                    self._update_corresponding_state(orig_descriptor)
                    self._mdib.descriptions.update_object_no_lock(orig_descriptor)
            for updates_dict, dest_list in ((self.alert_state_updates, proc.alert_updates),
                                            (self.metric_state_updates, proc.metric_updates),
                                            (self.context_state_updates, proc.ctxt_updates),
                                            (self.component_state_updates, proc.comp_updates),
                                            (self.operational_state_updates, proc.op_updates),
                                            (self.rt_sample_state_updates, proc.rt_updates),
                                            ):
                updates = self._handle_state_updates(updates_dict)
                dest_list.extend(updates)
        return proc

    def _update_corresponding_state(self, descriptor_container: AbstractDescriptorProtocol):
        updates_dict = self._get_states_update(descriptor_container)
        if descriptor_container.is_context_descriptor:
            all_context_states = self._mdib.context_states.descriptor_handle.get(
                descriptor_container.Handle, [])
            for context_state in all_context_states:
                state_update = updates_dict.get(context_state.Handle)
                if state_update is not None:
                    # the state has also been updated directly in transaction.
                    # update descriptor version
                    old_state, new_state = state_update.old, state_update.new
                else:
                    old_state = context_state
                    new_state = old_state.mk_copy()
                    updates_dict[context_state.Handle] = TransactionItem(old_state, new_state)
                new_state.descriptor_container = descriptor_container
                new_state.increment_state_version()
                new_state.update_descriptor_version()
        else:
            # check if state is already present in this transaction
            tr_item = updates_dict.get(descriptor_container.Handle)
            if tr_item is not None:
                # the state has also been updated directly in transaction.
                # update descriptor version
                if tr_item.new is None:
                    msg = f'State deleted? That should not be possible! handle = {descriptor_container.Handle}'
                    raise ValueError(msg)
                tr_item.new.update_descriptor_version()
            else:
                old_state = self._mdib.states.descriptor_handle.get_one(
                    descriptor_container.Handle, allow_none=True)
                if old_state is not None:
                    new_state = old_state.mk_copy()
                    new_state.descriptor_container = descriptor_container
                    new_state.DescriptorVersion = descriptor_container.DescriptorVersion
                    new_state.increment_state_version()
                    updates_dict[descriptor_container.Handle] = TransactionItem(old_state, new_state)

    def _increment_parent_descriptor_version(self, proc: TransactionResult,
                                             descriptor_container: AbstractDescriptorProtocol):
        parent_descriptor_container = self._mdib.descriptions.handle.get_one(
            descriptor_container.parent_handle, allow_none=True)
        if parent_descriptor_container is not None:
            parent_descriptor_container.increment_descriptor_version()
            proc.descr_updated.append(parent_descriptor_container.mk_copy())
            self._update_corresponding_state(parent_descriptor_container)

    def _get_states_update(self, container: AbstractStateProtocol | AbstractDescriptorProtocol) -> dict:
        if getattr(container, 'is_realtime_sample_array_metric_state', False) \
                or getattr(container, 'is_realtime_sample_array_metric_descriptor', False):
            return self.rt_sample_state_updates
        if getattr(container, 'is_metric_state', False) or getattr(container, 'is_metric_descriptor', False):
            return self.metric_state_updates
        if getattr(container, 'is_alert_state', False) or getattr(container, 'is_alert_descriptor', False):
            return self.alert_state_updates
        if getattr(container, 'is_component_state', False) or getattr(container, 'is_component_descriptor', False):
            return self.component_state_updates
        if getattr(container, 'is_operational_state', False) or getattr(container, 'is_operational_descriptor', False):
            return self.operational_state_updates
        if getattr(container, 'is_context_state', False) or getattr(container, 'is_context_descriptor', False):
            return self.context_state_updates
        msg = f'Unhandled case {container}'
        raise NotImplementedError(msg)


class StateTransactionBase(_TransactionBase):
    """Base Class for all transactions that modify states."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = {}  # will be set to proper value in derived classes

    def has_state(self, descriptor_handle: str) -> bool:
        """Check if transaction has a state with given handle."""
        return descriptor_handle in self._state_updates

    def unget_state(self, state_container: AbstractStateProtocol):
        """Forget a state that was provided before by a get_state or add_state call."""
        if state_container.DescriptorHandle in self._state_updates:
            del self._state_updates[state_container.DescriptorHandle]

    def get_state(self, descriptor_handle: str) -> AbstractStateProtocol:
        """Read a state from mdib and add it to the transaction.

        If the type of the state does not match the transaction type, an ApiUsageError is thrown.
        """
        if not descriptor_handle:
            raise ValueError('No handle for state specified')
        if descriptor_handle in self._state_updates:
            msg = f'State {descriptor_handle} already in updated set!'
            raise ValueError(msg)

        mdib_state = self._mdib.states.descriptor_handle.get_one(descriptor_handle, allow_none=False)
        if not self._is_correct_state_type(mdib_state):
            msg = f'Wrong data type in transaction! {self.__class__.__name__}, {mdib_state}'
            raise ApiUsageError(msg)

        copied_state = mdib_state.mk_copy()
        copied_state.increment_state_version()
        self._state_updates[descriptor_handle] = TransactionItem(mdib_state, copied_state)
        return copied_state

    def actual_descriptor(self, descriptor_handle: str) -> AbstractDescriptorProtocol:
        """Look descriptor in mdib, state transaction cannot have descriptor changes."""
        return self._mdib.descriptions.handle.get_one(descriptor_handle)

    def write_entity(self, entity: Entity, adjust_version_counter: bool = True):
        """Insert or update an entity."""
        if entity.is_multi_state:
            msg = f'Transaction {self.__class__.__name__} does not handle multi state entities!'
            raise ApiUsageError(msg)

        if not self._is_correct_state_type(entity.state):
            msg = f'Wrong data type in transaction! {self.__class__.__name__}, {entity.state}'
            raise ApiUsageError(msg)

        descriptor_handle = entity.state.DescriptorHandle
        old_state = self._mdib.states.descriptor_handle.get_one(entity.handle, allow_none=True)
        tmp_state = copy.deepcopy(entity.state)
        if adjust_version_counter:
            descriptor_container = self._mdib.descriptions.handle.get_one(descriptor_handle)
            tmp_state.DescriptorVersion = descriptor_container.DescriptorVersion
            if old_state is not None:
                # update from old state
                tmp_state.StateVersion = old_state.StateVersion + 1
            else:
                # new state, update version from saved versions in mdib if exists
                self._mdib.states.set_version(tmp_state)

        self._state_updates[descriptor_handle] = TransactionItem(old=old_state, new=tmp_state)

    def write_entities(self, entities: list[Entity | MultiStateEntity], adjust_version_counter: bool = True):
        """Write entities in order parents first."""
        for entity in entities:
            # check all states before writing any of them
            if entity.is_multi_state:
                msg = f'Transaction {self.__class__.__name__} does not handle multi state entities!'
                raise ApiUsageError(msg)

            if not self._is_correct_state_type(entity.state):
                msg = f'Wrong data type in transaction! {self.__class__.__name__}, {entity.state}'
                raise ApiUsageError(msg)
        for ent in entities:
            self.write_entity(ent, adjust_version_counter)

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:  # noqa: ARG004
        return False


class AlertStateTransaction(StateTransactionBase):
    """A Transaction for alert states."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = self.alert_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:
        """Process transaction and create a TransactionResult."""
        if set_determination_time:
            for tr_item in self._state_updates.values():
                new_state = tr_item.new
                old_state = tr_item.old
                if new_state is None or not hasattr(new_state, 'Presence'):
                    continue
                if old_state is None:
                    if new_state.Presence:
                        new_state.DeterminationTime = time.time()
                elif new_state.is_alert_condition and new_state.Presence != old_state.Presence:
                    new_state.DeterminationTime = time.time()
        proc = TransactionResult()
        if self._state_updates:
            self._mdib.mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.alert_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_alert_state


class MetricStateTransaction(StateTransactionBase):
    """A Transaction for metric states (except real time samples)."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = self.metric_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:
        """Process transaction and create a TransactionResult."""
        if set_determination_time:
            for tr_item in self._state_updates.values():
                if tr_item.new is not None and tr_item.new.MetricValue is not None:
                    tr_item.new.MetricValue.DeterminationTime = time.time()
        proc = TransactionResult()
        if self._state_updates:
            self._mdib.mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.metric_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_metric_state and not state.is_realtime_sample_array_metric_state


class ComponentStateTransaction(StateTransactionBase):
    """A Transaction for component states."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = self.component_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            self._mdib.mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.comp_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_component_state


class RtStateTransaction(StateTransactionBase):
    """A Transaction for real time sample states."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = self.rt_sample_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            self._mdib.mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.rt_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_realtime_sample_array_metric_state


class OperationalStateTransaction(StateTransactionBase):
    """A Transaction for operational states."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = self.operational_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            self._mdib.mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.op_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_operational_state


class ContextStateTransaction(_TransactionBase):
    """A Transaction for context states."""

    def __init__(self,
                 device_mdib_container: ProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(device_mdib_container, logger)
        self._state_updates = self.context_state_updates

    def get_context_state(self, context_state_handle: str) -> AbstractMultiStateProtocol:
        """Read a ContextState from mdib with given state handle."""
        if not context_state_handle:
            raise ValueError('No handle for context state specified')
        if context_state_handle in self._state_updates:
            msg = f'Context State {context_state_handle} already in updated set!'
            raise ValueError(msg)

        mdib_state = self._mdib.context_states.handle.get_one(context_state_handle, allow_none=False)
        copied_state = mdib_state.mk_copy()
        copied_state.increment_state_version()
        self._state_updates[context_state_handle] = TransactionItem(mdib_state, copied_state)
        return copied_state

    def mk_context_state(self, descriptor_handle: str,
                         context_state_handle: str | None = None,
                         adjust_state_version: bool = True,
                         set_associated: bool = False) -> AbstractMultiStateProtocol:
        """Create a new ContextStateContainer and add it to transaction.

        If context_state_handle is None, a unique handle is generated.
        """
        if not descriptor_handle:
            raise ValueError('No descriptor handle for context state specified')
        if context_state_handle in self._state_updates:
            msg = f'Context State {context_state_handle} already in updated set!'
            raise ValueError(msg)
        descriptor_container = self._mdib.descriptions.handle.get_one(descriptor_handle, allow_none=False)
        if not descriptor_container.is_context_descriptor:
            msg = f'Descriptor {descriptor_handle} is not a context descriptor!'
            raise ValueError(msg)

        if context_state_handle is not None:
            old_state_container = self._mdib.context_states.handle.get_one(context_state_handle, allow_none=True)
            if old_state_container is not None:
                msg = f'ContextState with handle={context_state_handle} already exists'
                raise ValueError(msg)

        new_state_container = self._mdib.data_model.mk_state_container(descriptor_container)
        new_state_container.Handle = context_state_handle or uuid.uuid4().hex
        if set_associated:
            # bind to new mdib version of this transaction
            new_state_container.BindingMdibVersion = self.new_mdib_version
            new_state_container.BindingStartTime = time.time()
            new_state_container.ContextAssociation = \
                self._mdib.data_model.pm_types.ContextAssociation.ASSOCIATED
        if context_state_handle is not None and adjust_state_version:
            self._mdib.context_states.set_version(new_state_container)

        self._state_updates[new_state_container.Handle] = TransactionItem(None, new_state_container)
        return cast(AbstractMultiStateProtocol, new_state_container)

    def add_state(self, state_container: AbstractMultiStateProtocol, adjust_state_version: bool = True):
        """Add a new context state to mdib."""
        if not state_container.is_context_state:
            # prevent this for simplicity reasons
            raise ApiUsageError('Transaction only handles context states!')

        if state_container.descriptor_container is None:
            descr = self._mdib.descriptions.handle.get_one(state_container.DescriptorHandle)
            state_container.descriptor_container = descr
            state_container.DescriptorVersion = state_container.descriptor_container.DescriptorVersion

        if adjust_state_version:
            self._mdib.context_states.set_version(state_container)
        self._state_updates[state_container.Handle] = TransactionItem(None, state_container)


    def disassociate_all(self,
                         context_descriptor_handle: str,
                         ignored_handle: str | None = None) -> list[str]:
        """Disassociate all associated states in mdib for context_descriptor_handle.

        The updated states are added to the transaction.
        The method returns a list of states that were disassociated.
        :param context_descriptor_handle: the handle of the context descriptor
        :param ignored_handle: the context state with this Handle shall not be touched.
        """
        pm_types = self._mdib.data_model.pm_types
        disassociated_state_handles = []
        old_state_containers = self._mdib.context_states.descriptor_handle.get(context_descriptor_handle, [])
        for old_state in old_state_containers:
            if old_state.Handle == ignored_handle or old_state.Handle in self._state_updates:
                # If state is already part of this transaction leave it also untouched, accept what the user wanted.
                continue
            if old_state.ContextAssociation != pm_types.ContextAssociation.DISASSOCIATED \
                    or old_state.UnbindingMdibVersion is None:
                self._logger.info('disassociate %s, handle=%s', old_state.NODETYPE.localname,
                                  old_state.Handle)
                transaction_state = self.get_context_state(old_state.Handle)
                transaction_state.ContextAssociation = pm_types.ContextAssociation.DISASSOCIATED
                if transaction_state.UnbindingMdibVersion is None:
                    transaction_state.UnbindingMdibVersion = self.new_mdib_version
                    transaction_state.BindingEndTime = time.time()
                disassociated_state_handles.append(transaction_state.Handle)
        return disassociated_state_handles

    def write_entity(self, entity: MultiStateEntity,
                  modified_handles: list[str],
                  adjust_version_counter: bool = True):
        """Insert or update a context state in mdib."""
        for handle in modified_handles:
            state_container = entity.states.get(handle)
            old_state = self._mdib.context_states.handle.get_one(handle, allow_none=True)
            if state_container is None:
                # a deleted state : this cannot be communicated via notification.
                # delete in internal_entity, and that is all
                if old_state is not None:
                    self._state_updates[handle] = TransactionItem(old=old_state, new=None)
                else:
                    msg = f'invalid handle {handle}!'
                    raise KeyError(msg)
                continue
            if not state_container.is_context_state:
                raise ApiUsageError('Transaction only handles context states!')

            tmp = copy.deepcopy(state_container)

            if old_state is None:
                # this is a new state
                tmp.descriptor_container = entity.descriptor
                tmp.DescriptorVersion = entity.descriptor.DescriptorVersion
                if adjust_version_counter:
                    self._mdib.context_states.set_version(tmp)
            elif adjust_version_counter:
                tmp.StateVersion = old_state.StateVersion + 1

            self._state_updates[state_container.Handle] = TransactionItem(old=old_state, new=tmp)

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            self._mdib.mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.ctxt_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_context_state


class TransactionResult:
    """The transaction result.

    Data is used to create notifications.
    """

    def __init__(self):
        # states and descriptors that were modified are stored here:
        self.descr_updated = []
        self.descr_created = []
        self.descr_deleted = []
        self.metric_updates = []
        self.alert_updates = []
        self.comp_updates = []
        self.ctxt_updates = []
        self.op_updates = []
        self.rt_updates = []

    @property
    def has_descriptor_updates(self) -> bool:
        """Return True if at least one descriptor is in result."""
        return len(self.descr_updated) > 0 or len(self.descr_created) > 0 or len(self.descr_deleted) > 0

    def all_states(self) -> list[AbstractStateProtocol]:
        """Return all states in this transaction."""
        return self.metric_updates + self.alert_updates + self.comp_updates + self.ctxt_updates \
               + self.op_updates + self.rt_updates


_transaction_type_lookup = {TransactionType.descriptor: DescriptorTransaction,
                            TransactionType.alert: AlertStateTransaction,
                            TransactionType.metric: MetricStateTransaction,
                            TransactionType.operational: OperationalStateTransaction,
                            TransactionType.context: ContextStateTransaction,
                            TransactionType.component: ComponentStateTransaction,
                            TransactionType.rt_sample: RtStateTransaction}


def mk_transaction(provider_mdib: ProviderMdib,
                   transaction_type: TransactionType,
                   logger: LoggerAdapter) -> AnyTransactionManagerProtocol:
    """Create a transaction according to transaction_type."""
    return _transaction_type_lookup[transaction_type](provider_mdib, logger)



# generated by "generate_qnames"
from sdc11073.namespaces import default_ns_helper

AbstractAlertDescriptor = default_ns_helper.PM.tag('AbstractAlertDescriptor')
AbstractAlertState = default_ns_helper.PM.tag('AbstractAlertState')
AbstractComplexDeviceComponentDescriptor = default_ns_helper.PM.tag('AbstractComplexDeviceComponentDescriptor')
AbstractComplexDeviceComponentState = default_ns_helper.PM.tag('AbstractComplexDeviceComponentState')
AbstractContextDescriptor = default_ns_helper.PM.tag('AbstractContextDescriptor')
AbstractContextState = default_ns_helper.PM.tag('AbstractContextState')
AbstractDescriptor = default_ns_helper.PM.tag('AbstractDescriptor')
AbstractDeviceComponentDescriptor = default_ns_helper.PM.tag('AbstractDeviceComponentDescriptor')
AbstractDeviceComponentState = default_ns_helper.PM.tag('AbstractDeviceComponentState')
AbstractMetricDescriptor = default_ns_helper.PM.tag('AbstractMetricDescriptor')
AbstractMetricState = default_ns_helper.PM.tag('AbstractMetricState')
AbstractMetricValue = default_ns_helper.PM.tag('AbstractMetricValue')
AbstractMultiState = default_ns_helper.PM.tag('AbstractMultiState')
AbstractOperationDescriptor = default_ns_helper.PM.tag('AbstractOperationDescriptor')
AbstractOperationState = default_ns_helper.PM.tag('AbstractOperationState')
AbstractSetStateOperationDescriptor = default_ns_helper.PM.tag('AbstractSetStateOperationDescriptor')
AbstractState = default_ns_helper.PM.tag('AbstractState')
AccessionIdentifier = default_ns_helper.PM.tag('AccessionIdentifier')
ActivateOperationDescriptor = default_ns_helper.PM.tag('ActivateOperationDescriptor')
ActivateOperationState = default_ns_helper.PM.tag('ActivateOperationState')
ActiveSyncProtocol = default_ns_helper.PM.tag('ActiveSyncProtocol')
AlertCondition = default_ns_helper.PM.tag('AlertCondition')
AlertConditionDescriptor = default_ns_helper.PM.tag('AlertConditionDescriptor')
AlertConditionState = default_ns_helper.PM.tag('AlertConditionState')
AlertSignal = default_ns_helper.PM.tag('AlertSignal')
AlertSignalDescriptor = default_ns_helper.PM.tag('AlertSignalDescriptor')
AlertSignalState = default_ns_helper.PM.tag('AlertSignalState')
AlertSystem = default_ns_helper.PM.tag('AlertSystem')
AlertSystemDescriptor = default_ns_helper.PM.tag('AlertSystemDescriptor')
AlertSystemState = default_ns_helper.PM.tag('AlertSystemState')
AllowedRange = default_ns_helper.PM.tag('AllowedRange')
AllowedValue = default_ns_helper.PM.tag('AllowedValue')
AllowedValues = default_ns_helper.PM.tag('AllowedValues')
Annotation = default_ns_helper.PM.tag('Annotation')
ApplyAnnotation = default_ns_helper.PM.tag('ApplyAnnotation')
ApprovedJurisdiction = default_ns_helper.PM.tag('ApprovedJurisdiction')
ApprovedJurisdictions = default_ns_helper.PM.tag('ApprovedJurisdictions')
Arg = default_ns_helper.PM.tag('Arg')
ArgName = default_ns_helper.PM.tag('ArgName')
Argument = default_ns_helper.PM.tag('Argument')
AssignedLocation = default_ns_helper.PM.tag('AssignedLocation')
BaseDemographics = default_ns_helper.PM.tag('BaseDemographics')
Battery = default_ns_helper.PM.tag('Battery')
BatteryDescriptor = default_ns_helper.PM.tag('BatteryDescriptor')
BatteryState = default_ns_helper.PM.tag('BatteryState')
BirthLength = default_ns_helper.PM.tag('BirthLength')
BirthWeight = default_ns_helper.PM.tag('BirthWeight')
Birthname = default_ns_helper.PM.tag('Birthname')
BodySite = default_ns_helper.PM.tag('BodySite')
CalibrationDocumentation = default_ns_helper.PM.tag('CalibrationDocumentation')
CalibrationInfo = default_ns_helper.PM.tag('CalibrationInfo')
CalibrationResult = default_ns_helper.PM.tag('CalibrationResult')
CapacityFullCharge = default_ns_helper.PM.tag('CapacityFullCharge')
CapacityRemaining = default_ns_helper.PM.tag('CapacityRemaining')
CapacitySpecified = default_ns_helper.PM.tag('CapacitySpecified')
Category = default_ns_helper.PM.tag('Category')
CauseInfo = default_ns_helper.PM.tag('CauseInfo')
Channel = default_ns_helper.PM.tag('Channel')
ChannelDescriptor = default_ns_helper.PM.tag('ChannelDescriptor')
ChannelState = default_ns_helper.PM.tag('ChannelState')
Characteristic = default_ns_helper.PM.tag('Characteristic')
ClinicalInfo = default_ns_helper.PM.tag('ClinicalInfo')
Clock = default_ns_helper.PM.tag('Clock')
ClockDescriptor = default_ns_helper.PM.tag('ClockDescriptor')
ClockState = default_ns_helper.PM.tag('ClockState')
Code = default_ns_helper.PM.tag('Code')
CodedValue = default_ns_helper.PM.tag('CodedValue')
CodingSystemName = default_ns_helper.PM.tag('CodingSystemName')
ComponentId = default_ns_helper.PM.tag('ComponentId')
ConceptDescription = default_ns_helper.PM.tag('ConceptDescription')
ContainmentTree = default_ns_helper.PM.tag('ContainmentTree')
ContainmentTreeEntry = default_ns_helper.PM.tag('ContainmentTreeEntry')
CoreData = default_ns_helper.PM.tag('CoreData')
Criticality = default_ns_helper.PM.tag('Criticality')
Current = default_ns_helper.PM.tag('Current')
DangerCode = default_ns_helper.PM.tag('DangerCode')
DateOfBirth = default_ns_helper.PM.tag('DateOfBirth')
Description = default_ns_helper.PM.tag('Description')
DeviceIdentifier = default_ns_helper.PM.tag('DeviceIdentifier')
DistributionRange = default_ns_helper.PM.tag('DistributionRange')
DistributionSampleArrayMetricDescriptor = default_ns_helper.PM.tag('DistributionSampleArrayMetricDescriptor')
DistributionSampleArrayMetricState = default_ns_helper.PM.tag('DistributionSampleArrayMetricState')
Documentation = default_ns_helper.PM.tag('Documentation')
DomainUnit = default_ns_helper.PM.tag('DomainUnit')
End = default_ns_helper.PM.tag('End')
EnsembleContext = default_ns_helper.PM.tag('EnsembleContext')
EnsembleContextDescriptor = default_ns_helper.PM.tag('EnsembleContextDescriptor')
EnsembleContextState = default_ns_helper.PM.tag('EnsembleContextState')
Entry = default_ns_helper.PM.tag('Entry')
EnumStringMetricDescriptor = default_ns_helper.PM.tag('EnumStringMetricDescriptor')
EnumStringMetricState = default_ns_helper.PM.tag('EnumStringMetricState')
ExpirationDate = default_ns_helper.PM.tag('ExpirationDate')
Familyname = default_ns_helper.PM.tag('Familyname')
FillerOrderNumber = default_ns_helper.PM.tag('FillerOrderNumber')
GestationalAge = default_ns_helper.PM.tag('GestationalAge')
Givenname = default_ns_helper.PM.tag('Givenname')
HeadCircumference = default_ns_helper.PM.tag('HeadCircumference')
Height = default_ns_helper.PM.tag('Height')
HumanReadableForm = default_ns_helper.PM.tag('HumanReadableForm')
Identification = default_ns_helper.PM.tag('Identification')
IdentifierName = default_ns_helper.PM.tag('IdentifierName')
ImagingProcedure = default_ns_helper.PM.tag('ImagingProcedure')
InstanceIdentifier = default_ns_helper.PM.tag('InstanceIdentifier')
Issuer = default_ns_helper.PM.tag('Issuer')
Jurisdiction = default_ns_helper.PM.tag('Jurisdiction')
Label = default_ns_helper.PM.tag('Label')
LimitAlertConditionDescriptor = default_ns_helper.PM.tag('LimitAlertConditionDescriptor')
LimitAlertConditionState = default_ns_helper.PM.tag('LimitAlertConditionState')
Limits = default_ns_helper.PM.tag('Limits')
LocalizedText = default_ns_helper.PM.tag('LocalizedText')
LocationContext = default_ns_helper.PM.tag('LocationContext')
LocationContextDescriptor = default_ns_helper.PM.tag('LocationContextDescriptor')
LocationContextState = default_ns_helper.PM.tag('LocationContextState')
LocationDetail = default_ns_helper.PM.tag('LocationDetail')
LocationReference = default_ns_helper.PM.tag('LocationReference')
LotNumber = default_ns_helper.PM.tag('LotNumber')
ManufactureDate = default_ns_helper.PM.tag('ManufactureDate')
Manufacturer = default_ns_helper.PM.tag('Manufacturer')
MaxLimits = default_ns_helper.PM.tag('MaxLimits')
MdDescription = default_ns_helper.PM.tag('MdDescription')
MdState = default_ns_helper.PM.tag('MdState')
Mdib = default_ns_helper.PM.tag('Mdib')
Mds = default_ns_helper.PM.tag('Mds')
MdsDescriptor = default_ns_helper.PM.tag('MdsDescriptor')
MdsState = default_ns_helper.PM.tag('MdsState')
Meaning = default_ns_helper.PM.tag('Meaning')
MeansContext = default_ns_helper.PM.tag('MeansContext')
MeansContextDescriptor = default_ns_helper.PM.tag('MeansContextDescriptor')
MeansContextState = default_ns_helper.PM.tag('MeansContextState')
Measurement = default_ns_helper.PM.tag('Measurement')
MeasurementUnit = default_ns_helper.PM.tag('MeasurementUnit')
MetaData = default_ns_helper.PM.tag('MetaData')
Metric = default_ns_helper.PM.tag('Metric')
MetricQuality = default_ns_helper.PM.tag('MetricQuality')
MetricValue = default_ns_helper.PM.tag('MetricValue')
Middlename = default_ns_helper.PM.tag('Middlename')
Modality = default_ns_helper.PM.tag('Modality')
ModelName = default_ns_helper.PM.tag('ModelName')
ModelNumber = default_ns_helper.PM.tag('ModelNumber')
ModifiableData = default_ns_helper.PM.tag('ModifiableData')
Mother = default_ns_helper.PM.tag('Mother')
Name = default_ns_helper.PM.tag('Name')
NeonatalPatientDemographicsCoreData = default_ns_helper.PM.tag('NeonatalPatientDemographicsCoreData')
NextCalibration = default_ns_helper.PM.tag('NextCalibration')
NumericMetricDescriptor = default_ns_helper.PM.tag('NumericMetricDescriptor')
NumericMetricState = default_ns_helper.PM.tag('NumericMetricState')
NumericMetricValue = default_ns_helper.PM.tag('NumericMetricValue')
OperatingJurisdiction = default_ns_helper.PM.tag('OperatingJurisdiction')
Operation = default_ns_helper.PM.tag('Operation')
OperationGroup = default_ns_helper.PM.tag('OperationGroup')
OperatorContext = default_ns_helper.PM.tag('OperatorContext')
OperatorContextDescriptor = default_ns_helper.PM.tag('OperatorContextDescriptor')
OperatorContextState = default_ns_helper.PM.tag('OperatorContextState')
OperatorDetails = default_ns_helper.PM.tag('OperatorDetails')
OrderDetail = default_ns_helper.PM.tag('OrderDetail')
Patient = default_ns_helper.PM.tag('Patient')
PatientContext = default_ns_helper.PM.tag('PatientContext')
PatientContextDescriptor = default_ns_helper.PM.tag('PatientContextDescriptor')
PatientContextState = default_ns_helper.PM.tag('PatientContextState')
PatientDemographicsCoreData = default_ns_helper.PM.tag('PatientDemographicsCoreData')
PatientType = default_ns_helper.PM.tag('PatientType')
PerformedOrderDetail = default_ns_helper.PM.tag('PerformedOrderDetail')
Performer = default_ns_helper.PM.tag('Performer')
PersonParticipation = default_ns_helper.PM.tag('PersonParticipation')
PersonReference = default_ns_helper.PM.tag('PersonReference')
PhysicalConnector = default_ns_helper.PM.tag('PhysicalConnector')
PhysicalConnectorInfo = default_ns_helper.PM.tag('PhysicalConnectorInfo')
PhysiologicalRange = default_ns_helper.PM.tag('PhysiologicalRange')
PlacerOrderNumber = default_ns_helper.PM.tag('PlacerOrderNumber')
ProductionSpec = default_ns_helper.PM.tag('ProductionSpec')
ProductionSpecification = default_ns_helper.PM.tag('ProductionSpecification')
ProtocolCode = default_ns_helper.PM.tag('ProtocolCode')
Race = default_ns_helper.PM.tag('Race')
Range = default_ns_helper.PM.tag('Range')
RealTimeSampleArrayMetricDescriptor = default_ns_helper.PM.tag('RealTimeSampleArrayMetricDescriptor')
RealTimeSampleArrayMetricState = default_ns_helper.PM.tag('RealTimeSampleArrayMetricState')
ReferenceRange = default_ns_helper.PM.tag('ReferenceRange')
ReferenceSource = default_ns_helper.PM.tag('ReferenceSource')
ReferringPhysician = default_ns_helper.PM.tag('ReferringPhysician')
RelatedMeasurement = default_ns_helper.PM.tag('RelatedMeasurement')
Relation = default_ns_helper.PM.tag('Relation')
RelevantClinicalInfo = default_ns_helper.PM.tag('RelevantClinicalInfo')
RemainingBatteryTime = default_ns_helper.PM.tag('RemainingBatteryTime')
RemedyInfo = default_ns_helper.PM.tag('RemedyInfo')
RequestedOrderDetail = default_ns_helper.PM.tag('RequestedOrderDetail')
RequestedProcedureId = default_ns_helper.PM.tag('RequestedProcedureId')
RequestingPhysician = default_ns_helper.PM.tag('RequestingPhysician')
ResultingClinicalInfo = default_ns_helper.PM.tag('ResultingClinicalInfo')
Role = default_ns_helper.PM.tag('Role')
SampleArrayValue = default_ns_helper.PM.tag('SampleArrayValue')
ScheduledProcedureStepId = default_ns_helper.PM.tag('ScheduledProcedureStepId')
Sco = default_ns_helper.PM.tag('Sco')
ScoDescriptor = default_ns_helper.PM.tag('ScoDescriptor')
ScoState = default_ns_helper.PM.tag('ScoState')
SerialNumber = default_ns_helper.PM.tag('SerialNumber')
Service = default_ns_helper.PM.tag('Service')
SetAlertStateOperationDescriptor = default_ns_helper.PM.tag('SetAlertStateOperationDescriptor')
SetAlertStateOperationState = default_ns_helper.PM.tag('SetAlertStateOperationState')
SetComponentStateOperationDescriptor = default_ns_helper.PM.tag('SetComponentStateOperationDescriptor')
SetComponentStateOperationState = default_ns_helper.PM.tag('SetComponentStateOperationState')
SetContextStateOperationDescriptor = default_ns_helper.PM.tag('SetContextStateOperationDescriptor')
SetContextStateOperationState = default_ns_helper.PM.tag('SetContextStateOperationState')
SetMetricStateOperationDescriptor = default_ns_helper.PM.tag('SetMetricStateOperationDescriptor')
SetMetricStateOperationState = default_ns_helper.PM.tag('SetMetricStateOperationState')
SetStringOperationDescriptor = default_ns_helper.PM.tag('SetStringOperationDescriptor')
SetStringOperationState = default_ns_helper.PM.tag('SetStringOperationState')
SetValueOperationDescriptor = default_ns_helper.PM.tag('SetValueOperationDescriptor')
SetValueOperationState = default_ns_helper.PM.tag('SetValueOperationState')
Sex = default_ns_helper.PM.tag('Sex')
Source = default_ns_helper.PM.tag('Source')
SpecType = default_ns_helper.PM.tag('SpecType')
Start = default_ns_helper.PM.tag('Start')
State = default_ns_helper.PM.tag('State')
StringMetricDescriptor = default_ns_helper.PM.tag('StringMetricDescriptor')
StringMetricState = default_ns_helper.PM.tag('StringMetricState')
StringMetricValue = default_ns_helper.PM.tag('StringMetricValue')
StudyInstanceUid = default_ns_helper.PM.tag('StudyInstanceUid')
SystemContext = default_ns_helper.PM.tag('SystemContext')
SystemContextDescriptor = default_ns_helper.PM.tag('SystemContextDescriptor')
SystemContextState = default_ns_helper.PM.tag('SystemContextState')
SystemSignalActivation = default_ns_helper.PM.tag('SystemSignalActivation')
TechnicalRange = default_ns_helper.PM.tag('TechnicalRange')
Temperature = default_ns_helper.PM.tag('Temperature')
TimeProtocol = default_ns_helper.PM.tag('TimeProtocol')
Title = default_ns_helper.PM.tag('Title')
Translation = default_ns_helper.PM.tag('Translation')
Type = default_ns_helper.PM.tag('Type')
Udi = default_ns_helper.PM.tag('Udi')
Unit = default_ns_helper.PM.tag('Unit')
Validator = default_ns_helper.PM.tag('Validator')
Value = default_ns_helper.PM.tag('Value')
VisitNumber = default_ns_helper.PM.tag('VisitNumber')
Vmd = default_ns_helper.PM.tag('Vmd')
VmdDescriptor = default_ns_helper.PM.tag('VmdDescriptor')
VmdState = default_ns_helper.PM.tag('VmdState')
Voltage = default_ns_helper.PM.tag('Voltage')
VoltageSpecified = default_ns_helper.PM.tag('VoltageSpecified')
Weight = default_ns_helper.PM.tag('Weight')
WorkflowContext = default_ns_helper.PM.tag('WorkflowContext')
WorkflowContextDescriptor = default_ns_helper.PM.tag('WorkflowContextDescriptor')
WorkflowContextState = default_ns_helper.PM.tag('WorkflowContextState')
WorkflowDetail = default_ns_helper.PM.tag('WorkflowDetail')



"""Implementation of alarm provider functionality."""
from __future__ import annotations

import time
import traceback
from threading import Event, Thread
from typing import TYPE_CHECKING, cast

from sdc11073.mdib.descriptorcontainers import AbstractSetStateOperationDescriptorContainer
from sdc11073.mdib.statecontainers import AbstractStateProtocol, AlertConditionStateContainer
from sdc11073.provider.operations import ExecuteResult

from . import providerbase

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.entityprotocol import EntityProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.mdib.transactionsprotocol import StateTransactionManagerProtocol
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase, OperationDefinitionProtocol
    from sdc11073.provider.sco import AbstractScoOperationsRegistry

    from .providerbase import OperationClassGetter


class AlertDelegateProvider(providerbase.ProviderRole):
    """Support alert delegation acc. to BICEPS chapter 6.2."""

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        super().__init__(mdib, log_prefix)

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Return a callable for this operation or None.

        Creates operation handler for:
        - set alert signal state
            => SetAlertStateOperation
                operation target Is an AlertSignalDescriptor
            handler = self._delegate_alert_signal
        """
        pm_names = self._mdib.data_model.pm_names
        op_target_handle = operation_descriptor_container.OperationTarget
        op_target_entity = self._mdib.entities.by_handle(op_target_handle)
        if pm_names.SetAlertStateOperationDescriptor == operation_descriptor_container.NODETYPE:
            if (pm_names.AlertSignalDescriptor == op_target_entity.node_type
                    and op_target_entity.descriptor.SignalDelegationSupported):
                # operation_descriptor_container is a SetAlertStateOperationDescriptor
                set_state_descriptor_container = cast(AbstractSetStateOperationDescriptorContainer,
                                                      operation_descriptor_container)
                modifiable_data = set_state_descriptor_container.ModifiableData
                if 'Presence' in modifiable_data \
                        and 'ActivationState' in modifiable_data \
                        and 'ActualSignalGenerationDelay' in modifiable_data:
                    operation = self._mk_operation_from_operation_descriptor(
                        operation_descriptor_container,
                        operation_cls_getter,
                        operation_handler=self._delegate_alert_signal,
                        timeout_handler=self._on_timeout_delegate_alert_signal)

                    self._logger.debug('%s: added handler "self._delegate_alert_signal" for %s target=%s',
                                       self.__class__.__name__, operation_descriptor_container, op_target_handle)
                    return operation

        return None  # None == no handler for this operation instantiated

    def _delegate_alert_signal(self, params: ExecuteParameters) -> ExecuteResult:
        """Handle operation call from remote (ExecuteHandler).

        Sets ActivationState, Presence and ActualSignalGenerationDelay of the corresponding state in mdib.
        If this is a delegable signal, it also sets the ActivationState of the fallback signal.
        """
        value = params.operation_request.argument
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names
        all_alert_signal_entities = self._mdib.entities.by_node_type(pm_names.AlertSignalDescriptor)

        operation_target_handle = params.operation_instance.operation_target_handle
        op_target_entity = self._mdib.entities.by_handle(operation_target_handle)

        self._logger.info('delegate alert signal %s of %s from %s to %s', operation_target_handle,
                          op_target_entity.state, op_target_entity.state.ActivationState, value.ActivationState)
        for name in params.operation_instance.descriptor_container.ModifiableData:
            tmp = getattr(value, name)
            setattr(op_target_entity.state, name, tmp)
        modified = []
        if op_target_entity.descriptor.SignalDelegationSupported:
            if value.ActivationState == pm_types.AlertActivation.ON:
                modified = self._pause_fallback_alert_signals(op_target_entity,
                                                              all_alert_signal_entities)
            else:
                modified = self._activate_fallback_alert_signals(op_target_entity,
                                                                 all_alert_signal_entities)
        with self._mdib.alert_state_transaction() as mgr:
            mgr.write_entity(op_target_entity)
            mgr.write_entities(modified)

        return ExecuteResult(operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def _on_timeout_delegate_alert_signal(self, operation_instance: OperationDefinitionProtocol):
        """TimeoutHandler for delegated signal."""
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names

        operation_target_handle = operation_instance.operation_target_handle
        op_target_entity = self._mdib.entities.by_handle(operation_target_handle)

        all_alert_signal_entities = self._mdib.entities.by_node_type(pm_names.AlertSignalDescriptor)
        self._logger.info('timeout alert signal delegate operation=%s target=%s',
                          operation_instance.handle, operation_target_handle)
        op_target_entity.state.ActivationState = pm_types.AlertActivation.OFF
        modified = self._activate_fallback_alert_signals(op_target_entity,
                                                         all_alert_signal_entities)

        with self._mdib.alert_state_transaction() as mgr:
            mgr.write_entity(op_target_entity)
            mgr.write_entities(modified)

    def _pause_fallback_alert_signals(self,
                                      delegable_signal_entity: EntityProtocol,
                                      all_signal_entities: list[EntityProtocol],
                                      ) -> list[EntityProtocol]:
        """Pause fallback signals.

        The idea of the fallback signal is to set it paused when the delegable signal is currently ON,
        and to set it back to ON when the delegable signal is not ON.
        This method sets the fallback to PAUSED value.
        :param delegable_signal_entity: the signal that the fallback signals are looked for.
        :param all_signal_entities: list of all signals
        :return: list of modified entities
        """
        pm_types = self._mdib.data_model.pm_types
        modified: list[EntityProtocol] = []
        # look for local fallback signal (same Manifestation), and set it to paused
        for fallback_entity in self._get_fallback_signals(delegable_signal_entity,
                                                          all_signal_entities):
            if fallback_entity.state.ActivationState != pm_types.AlertActivation.PAUSED:
                fallback_entity.state.ActivationState = pm_types.AlertActivation.PAUSED
                modified.append(fallback_entity)
        return modified

    def _activate_fallback_alert_signals(self, delegable_signal_entity: EntityProtocol,
                                         all_signal_entities: list[EntityProtocol],
                                         ) -> list[EntityProtocol]:
        pm_types = self._mdib.data_model.pm_types
        modified: list[EntityProtocol] = []

        # look for local fallback signal (same Manifestation), and set it to paused
        for fallback_entity in self._get_fallback_signals(delegable_signal_entity,
                                                          all_signal_entities):
            if fallback_entity.state.ActivationState == pm_types.AlertActivation.PAUSED:
                fallback_entity.state.ActivationState = pm_types.AlertActivation.ON
                modified.append(fallback_entity)
        return modified

    @staticmethod
    def _get_fallback_signals(delegable_signal_entity: EntityProtocol,
                              all_signal_entities: list[EntityProtocol]) -> list[EntityProtocol]:
        """Return a list of all fallback signals for descriptor.

        looks in all_signal_descriptors for a signal with same ConditionSignaled and same
        Manifestation as delegable_signal_descriptor and SignalDelegationSupported == True.
        """
        return [tmp for tmp in all_signal_entities if not tmp.descriptor.SignalDelegationSupported
                and tmp.descriptor.Manifestation == delegable_signal_entity.descriptor.Manifestation
                and tmp.descriptor.ConditionSignaled == delegable_signal_entity.descriptor.ConditionSignaled]


class AlertSystemStateMaintainer(providerbase.ProviderRole):
    """Provide some generic alarm handling functionality.

    - runs periodic job to update currently present alarms in AlertSystemState
    """

    WORKER_THREAD_INTERVAL = 1.0  # seconds
    self_check_safety_margin = 1.0  # how many seconds before SelfCheckInterval elapses a new self check is performed.

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        super().__init__(mdib, log_prefix)

        self._stop_worker = Event()
        self._worker_thread = None

    def init_operations(self, sco: AbstractScoOperationsRegistry):
        """Start a worker thread that periodically updates AlertSystemStateContainers."""
        super().init_operations(sco)
        self._worker_thread = Thread(target=self._worker_thread_loop)
        self._worker_thread.daemon = True
        self._worker_thread.start()

    def stop(self):
        """Stop worker thread."""
        self._stop_worker.set()
        self._worker_thread.join()

    def _worker_thread_loop(self):
        # delay start of operation
        shall_stop = self._stop_worker.wait(timeout=self.WORKER_THREAD_INTERVAL)
        if shall_stop:
            return

        while True:
            shall_stop = self._stop_worker.wait(timeout=self.WORKER_THREAD_INTERVAL)
            if shall_stop:
                return
            self._update_alert_system_state_current_alerts()

    def _update_alert_system_state_current_alerts(self):
        """Update AlertSystemState present alarms list."""
        try:
            entities_needing_update = self._get_alert_system_entities_needing_update()
            if len(entities_needing_update) > 0:
                with self._mdib.alert_state_transaction() as mgr:
                    self._update_alert_system_states(entities_needing_update)
                    mgr.write_entities(entities_needing_update)
        except Exception: # noqa: BLE001
            self._logger.error('_update_alert_system_state_current_alerts: %s', traceback.format_exc())  # noqa: TRY400

    def _get_alert_system_entities_needing_update(self) -> list[EntityProtocol]:
        pm_names = self._mdib.data_model.pm_names
        entities_needing_update = []
        try:
            all_alert_system_entities = self._mdib.entities.by_node_type(pm_names.AlertSystemDescriptor)
            for alert_system_entity in all_alert_system_entities:
                if alert_system_entity.state is not None:
                    self_check_period = alert_system_entity.descriptor.SelfCheckPeriod
                    if self_check_period is not None:
                        last_self_check = alert_system_entity.state.LastSelfCheck or 0.0
                        if time.time() - last_self_check >= self_check_period - self.self_check_safety_margin:
                            entities_needing_update.append(alert_system_entity)
        except Exception: # noqa: BLE001
            self._logger.error('_get_alert_system_entities_needing_update: %s', traceback.format_exc())  # noqa: TRY400
        return entities_needing_update

    def _update_alert_system_states(self, alert_system_entities: Iterable[EntityProtocol]):
        """Update alert system states."""
        pm_types = self._mdib.data_model.pm_types

        for alert_system_entity in alert_system_entities:
            all_child_entities = self._mdib.entities.by_parent_handle(alert_system_entity.handle)
            all_alert_condition_entities = [d for d in all_child_entities if d.descriptor.is_alert_condition_descriptor]
            # select all state containers with technical alarms present
            all_tech_entities = [d for d in all_alert_condition_entities if
                                 d.descriptor.Kind == pm_types.AlertConditionKind.TECHNICAL]
            all_present_tech_entities = [s for s in all_tech_entities if s.state.Presence]
            # select all state containers with physiological alarms present
            all_phys_entities = [d for d in all_alert_condition_entities if
                                 d.descriptor.Kind == pm_types.AlertConditionKind.PHYSIOLOGICAL]
            all_present_phys_entities = [s for s in all_phys_entities if s.state.Presence]

            alert_system_entity.state.PresentTechnicalAlarmConditions = [e.handle for e in all_present_tech_entities]
            alert_system_entity.state.PresentPhysiologicalAlarmConditions = [e.handle for e in
                                                                             all_present_phys_entities]
            alert_system_entity.state.LastSelfCheck = time.time()
            alert_system_entity.state.SelfCheckCount = 1 if alert_system_entity.state.SelfCheckCount is None \
                else alert_system_entity.state.SelfCheckCount + 1


class AlertPreCommitHandler(providerbase.ProviderRole):
    """Provide some generic alarm handling functionality.

    - in pre commit handler it updates present alarms list of alarm system states
    """

    def on_pre_commit(self, mdib: ProviderMdibProtocol, transaction: StateTransactionManagerProtocol):
        """Manipulate the transaction.

        - Updates alert system states and adds them to transaction, if at least one of its alert
          conditions changed ( is in transaction).
        - Updates all AlertSignals for changed Alert Conditions and adds them to transaction.

        :param mdib:
        :param transaction:
        :return:
        """
        if not transaction.alert_state_updates:
            # nothing to do
            return

        all_alert_signal_entities = self._mdib.entities.by_node_type(
            self._mdib.data_model.pm_names.AlertSignalDescriptor)
        changed_alert_condition_states = self._get_changed_alert_condition_states(transaction)
        # change AlertSignal Settings in order to be compliant with changed Alert Conditions
        for changed_alert_condition_state in changed_alert_condition_states:
            self._update_alert_signals(changed_alert_condition_state,
                                       all_alert_signal_entities,
                                       mdib,
                                       transaction)

        # find all alert systems for changed_alert_condition_states
        alert_system_states = self._find_alert_systems_with_modifications(transaction,
                                                                          changed_alert_condition_states)
        if alert_system_states:
            # add found alert system states to transaction
            self._update_alert_system_states(mdib, alert_system_states, transaction)

    @staticmethod
    def _update_alert_system_states(mdib: ProviderMdibProtocol,
                                    alert_system_states: Iterable[AbstractStateProtocol],
                                    transaction: StateTransactionManagerProtocol):
        """Update alert system states PresentTechnicalAlarmConditions and PresentPhysiologicalAlarmConditions."""
        pm_types = mdib.data_model.pm_types

        def _get_alert_state(state: AbstractStateProtocol) -> AbstractStateProtocol:
            """Return the equivalent state from current transaction, if it already in transaction."""
            _item = transaction.get_state_transaction_item(state.DescriptorHandle)
            if _item is not None:
                return _item.new
            return state

        for _alert_system_state in alert_system_states:
            write_entity = True
            tr_item = transaction.get_state_transaction_item(_alert_system_state.DescriptorHandle)
            # If the alert system state is already part of the transaction, make changes in that instance instead.
            if tr_item is not None:
                alert_system_state = tr_item.new
                write_entity = False
            else:
                alert_system_state = _alert_system_state
            all_child_entities = mdib.entities.by_parent_handle(alert_system_state.DescriptorHandle)
            all_alert_condition_entities = [d for d in all_child_entities if d.descriptor.is_alert_condition_descriptor]
            # select all state containers with technical alarms present
            all_tech_entities = [d for d in all_alert_condition_entities if
                                 d.descriptor.Kind == pm_types.AlertConditionKind.TECHNICAL]
            _all_tech_states = [_get_alert_state(d.state) for d in all_tech_entities]
            all_tech_states = cast(list[AlertConditionStateContainer], _all_tech_states)
            all_present_tech_states = [s for s in all_tech_states if s.Presence]

            all_phys_entities = [d for d in all_alert_condition_entities if
                                 d.descriptor.Kind == pm_types.AlertConditionKind.PHYSIOLOGICAL]
            _all_phys_states = [_get_alert_state(d.state) for d in all_phys_entities]
            all_phys_states = cast(list[AlertConditionStateContainer], _all_phys_states)
            all_phys_states = [s for s in all_phys_states if s is not None]
            all_present_phys_states = [s for s in all_phys_states if s.Presence]

            alert_system_state.PresentTechnicalAlarmConditions = [s.DescriptorHandle for s in all_present_tech_states]
            alert_system_state.PresentPhysiologicalAlarmConditions = [s.DescriptorHandle for s in
                                                                      all_present_phys_states]
            if write_entity:
                transaction.write_entity(alert_system_state)

    @staticmethod
    def _get_changed_alert_condition_states(transaction: StateTransactionManagerProtocol) -> list[
        AbstractStateProtocol]:
        """Return all alert conditions in current transaction."""
        result = []
        for item in list(transaction.alert_state_updates.values()):
            tmp = item.old if item.new is None else item.new
            if tmp.is_alert_condition:
                result.append(tmp)
        return result

    @staticmethod
    def _update_alert_signals(changed_alert_condition: AbstractStateProtocol,
                              all_alert_signal_entities: list[EntityProtocol],
                              mdib: ProviderMdibProtocol,
                              transaction: StateTransactionManagerProtocol):
        """Handle alert signals for a changed alert condition.

        This method only changes states of local signals.
        Handling of delegated signals is in the responsibility of the delegated device!
        """
        pm_types = mdib.data_model.pm_types

        my_alert_signal_entities = [e for e in all_alert_signal_entities
                                    if e.descriptor.ConditionSignaled == changed_alert_condition.DescriptorHandle]
        # separate remote from local
        remote_alert_signal_entities = [a for a in my_alert_signal_entities if a.descriptor.SignalDelegationSupported]
        local_alert_signal_entities = [a for a in my_alert_signal_entities if
                                       not a.descriptor.SignalDelegationSupported]

        # look for active delegations (we only need the Manifestation value here)
        active_delegate_manifestations = []
        for entity in remote_alert_signal_entities:
            if (entity.state.Presence != pm_types.AlertSignalPresence.OFF
                    and entity.state.Location == pm_types.AlertSignalPrimaryLocation.REMOTE):
                active_delegate_manifestations.append(entity.descriptor.Manifestation) # noqa: PERF401

        # this lookup gives the values that a local signal shall have:
        # key = (Cond.Presence, isDelegated): value = (SignalState.ActivationState, SignalState.Presence)
        # see BICEPS standard table 9: valid combinations of alert activation states, alert condition presence, ...
        # this is the relevant subset for our case
        lookup = {(True, True): (pm_types.AlertActivation.PAUSED, pm_types.AlertSignalPresence.OFF),
                  (True, False): (pm_types.AlertActivation.ON, pm_types.AlertSignalPresence.ON),
                  (False, True): (pm_types.AlertActivation.PAUSED, pm_types.AlertSignalPresence.OFF),
                  (False, False): (pm_types.AlertActivation.ON, pm_types.AlertSignalPresence.OFF),
                  }
        for entity in local_alert_signal_entities:
            tr_item = transaction.get_state_transaction_item(entity.handle)
            if tr_item is None:
                # is this local signal delegated?
                is_delegated = entity.descriptor.Manifestation in active_delegate_manifestations
                activation, presence = lookup[(changed_alert_condition.Presence, is_delegated)]

                if entity.state.ActivationState != activation or entity.state.Presence != presence:
                    entity.state.ActivationState = activation
                    entity.state.Presence = presence
                    transaction.write_entity(entity)

    def _find_alert_systems_with_modifications(self,
                                               transaction: StateTransactionManagerProtocol,
                                               changed_alert_conditions: list[AbstractStateProtocol]) \
            -> set[AbstractStateProtocol]:
        # find all alert systems for the changed alert conditions
        alert_system_states = set()
        for tmp in changed_alert_conditions:
            alert_condition_entity = self._mdib.entities.by_handle(tmp.DescriptorHandle)
            alert_system_entity = self._mdib.entities.by_handle(alert_condition_entity.parent_handle)

            if alert_system_entity.handle not in transaction.alert_state_updates:
                transaction.write_entity(alert_system_entity)

            transaction_item = transaction.alert_state_updates[alert_system_entity.handle]
            if transaction_item.new is not None:
                alert_system_states.add(transaction_item.new)
        return alert_system_states



"""Module containing utilities and helper methods regarding xml."""
from __future__ import annotations

import copy
import sys
from typing import Callable

from lxml import etree

if sys.version_info >= (3, 10):
    from typing import TypeAlias

    LxmlElement: TypeAlias = etree._Element  # noqa: SLF001
else:
    from typing_extensions import TypeAlias

    LxmlElement: TypeAlias = etree._Element  # noqa: SLF001


def copy_element(node: LxmlElement, method: Callable[[LxmlElement], LxmlElement] = copy.deepcopy) -> LxmlElement:
    """Copy and preserve complete namespace.

    :param node: node to be copied
    :param method: method that creates a duplication of the root node
    :return: new node
    """
    # walk from target to root
    current = node
    ns_map_list: list[dict[str, str]] = []  # saves all namespaces
    while current is not None:
        ns_map_list.append({k: v for k, v in current.nsmap.items() if k})  # filter for default namespace
        current = current.getparent()

    # create new instance
    root_tree = node.getroottree()
    current = method(root_tree.getroot())
    x_path_steps = root_tree.getpath(node).split('/')[1:]
    assert len(x_path_steps) == len(ns_map_list)

    # walk from root to target
    ns_map_list.reverse()
    for i, step in enumerate(x_path_steps):
        x_path_elements = current.xpath(f'/{step}' if i == 0 else step, namespaces=ns_map_list[i])
        assert len(x_path_elements) == 1
        current = x_path_elements[0]
    return current


def copy_node_wo_parent(node: LxmlElement, method: Callable[[LxmlElement], LxmlElement] = copy.deepcopy) -> LxmlElement:
    """Copy node but only keep relevant information and no parent.

    :param node: node to be copied
    :param method: method that copies an etree element
    :return: new node
    """
    new_node = etree.Element(node.tag, attrib=node.attrib, nsmap=node.nsmap)
    new_node.text = node.text
    new_node.tail = node.tail
    new_node.extend(method(child) for child in node)
    return new_node


class QName(etree.QName):
    """Implements copy and deepcopy for lxml QName as it is unpickleable."""

    def __copy__(self):
        return QName(self.text)

    def __deepcopy__(self, _: dict):
        return QName(self.text)



from . import xml_structure as cp
from .dpws_types import HostServiceType, HostedServiceType
from .dpws_types import ThisDeviceType, ThisModelType, DeviceMetadataDialectURI, DeviceRelationshipTypeURI
from .msg_types import MessageType
from .pm_types import PropertyBasedPMType
from sdc11073.namespaces import default_ns_helper

########## Meta Data Exchange #########
wsx_tag = default_ns_helper.WSX.tag  # shortcut
dpws_tag = default_ns_helper.DPWS.tag  # shortcut
wsa_tag = default_ns_helper.WSA.tag  # shortcut


class GetMetadata(MessageType):
    NODETYPE = wsx_tag('GetMetadata')
    action = 'http://schemas.xmlsoap.org/ws/2004/09/mex/GetMetadata/Request'
    Dialect = cp.AnyUriTextElement(wsx_tag('Dialect'), is_optional=True)
    Identifier = cp.AnyUriTextElement(wsx_tag('Identifier'), is_optional=True)
    _props = ('Dialect', 'Identifier')


class ThisModelMetadataSection(PropertyBasedPMType):
    MetadataReference = cp.SubElementProperty(dpws_tag('ThisModel'),
                                              value_class=ThisModelType)
    Location = cp.AnyUriTextElement(wsx_tag('Location'), is_optional=True)
    Dialect = cp.AnyURIAttributeProperty('Dialect',
                                         default_py_value=DeviceMetadataDialectURI.THIS_MODEL,
                                         is_optional=False)
    Identifier = cp.AnyURIAttributeProperty('Identifier')
    _props = ('MetadataReference', 'Location', 'Dialect', 'Identifier')


class ThisDeviceMetadataSection(PropertyBasedPMType):
    MetadataReference = cp.SubElementProperty(dpws_tag('ThisDevice'),
                                              value_class=ThisDeviceType)
    Location = cp.AnyUriTextElement(wsx_tag('Location'), is_optional=True)
    Dialect = cp.AnyURIAttributeProperty('Dialect',
                                         default_py_value=DeviceMetadataDialectURI.THIS_DEVICE,
                                         is_optional=False)
    Identifier = cp.AnyURIAttributeProperty('Identifier')
    _props = ('MetadataReference', 'Location', 'Dialect', 'Identifier')


class MetaDataRelationship(PropertyBasedPMType):
    Type = cp.AnyURIAttributeProperty(attribute_name='Type', default_py_value=DeviceRelationshipTypeURI.HOST)
    Host = cp.SubElementProperty(dpws_tag('Host'), value_class=HostServiceType)
    Hosted = cp.SubElementListProperty(dpws_tag('Hosted'), value_class=HostedServiceType)
    _props = ('Type', 'Host', 'Hosted')


class RelationshipMetadataSection(PropertyBasedPMType):
    MetadataReference = cp.SubElementProperty(dpws_tag('Relationship'),
                                              value_class=MetaDataRelationship,
                                              default_py_value=MetaDataRelationship())
    Location = cp.AnyUriTextElement(wsx_tag('Location'), is_optional=True)
    Dialect = cp.AnyURIAttributeProperty('Dialect',
                                         default_py_value=DeviceMetadataDialectURI.RELATIONSHIP,
                                         is_optional=False)
    Identifier = cp.AnyURIAttributeProperty('Identifier')
    _props = ('MetadataReference', 'Location', 'Dialect', 'Identifier')


class LocationMetadataSection(PropertyBasedPMType):
    Location = cp.AnyUriTextElement(wsx_tag('Location'), is_optional=True)
    Dialect = cp.AnyURIAttributeProperty('Dialect',
                                         default_py_value=default_ns_helper.WSDL.namespace,
                                         is_optional=False)
    Identifier = cp.AnyURIAttributeProperty('Identifier')
    _props = ('Location', 'Dialect', 'Identifier')


dialect_lookup = {DeviceMetadataDialectURI.THIS_MODEL: (ThisModelMetadataSection,
                                                        'this_model',
                                                        'MetadataReference'),
                  DeviceMetadataDialectURI.THIS_DEVICE: (ThisDeviceMetadataSection,
                                                         'this_device',
                                                         'MetadataReference'),
                  DeviceMetadataDialectURI.RELATIONSHIP: (RelationshipMetadataSection,
                                                          'relationship',
                                                          'MetadataReference'),
                  default_ns_helper.WSDL.namespace: (LocationMetadataSection,
                                                     'wsdl_location',
                                                     'Location')
                  }


class Metadata(MessageType):
    NODETYPE = wsx_tag('Metadata')
    action = f'{default_ns_helper.WXF.namespace}/GetResponse'
    MetadataSection = cp.SubElementListProperty(wsx_tag('MetadataSection'),
                                                value_class=PropertyBasedPMType)
    _props = ('MetadataSection',)

    def __init__(self):
        super().__init__()
        self.this_model = None
        self.this_device = None
        self.wsdl_location = None
        self.relationship = None

    @classmethod
    def from_node(cls, node):
        """ default from_node Constructor that provides no arguments for class __init__"""
        meta_data = cls()
        for section_node in node[0]:
            dialect = section_node.attrib.get('Dialect')
            section_cls, member_name, section_member_name = dialect_lookup.get(dialect)
            if section_cls is not None:
                section = section_cls.from_node(section_node)
                meta_data.MetadataSection.append(section)
                setattr(meta_data, member_name, getattr(section, section_member_name))
            else:
                print(f'unknown dialect {dialect}')
        return meta_data



"""The module contains extensions to the functionality of the EntityConsumerMdib."""
from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073 import observableproperties as properties

if TYPE_CHECKING:
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.pysoap.msgreader import ReceivedMessage

    from .entity_consumermdib import EntityConsumerMdib

class EntityConsumerMdibMethods:
    """Extra methods for consumer mdib that are not part of core functionality."""

    def __init__(self, consumer_mdib: EntityConsumerMdib, logger: LoggerAdapter):
        self._mdib = consumer_mdib
        self._sdc_client = consumer_mdib.sdc_client
        self._msg_reader = self._sdc_client.msg_reader
        self._logger = logger

    def bind_to_client_observables(self):
        """Connect the mdib with the notifications from consumer."""
        properties.bind(self._sdc_client, waveform_report=self._on_waveform_report)
        properties.bind(self._sdc_client, episodic_metric_report=self._on_episodic_metric_report)
        properties.bind(self._sdc_client, episodic_alert_report=self._on_episodic_alert_report)
        properties.bind(self._sdc_client, episodic_context_report=self._on_episodic_context_report)
        properties.bind(self._sdc_client, episodic_component_report=self._on_episodic_component_report)
        properties.bind(self._sdc_client, description_modification_report=self._on_description_modification_report)
        properties.bind(self._sdc_client, episodic_operational_state_report=self._on_operational_state_report)

    def _on_episodic_metric_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_episodic_metric_report')
        self._mdib.process_incoming_metric_states_report(received_message_data)

    def _on_episodic_alert_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_episodic_alert_report')
        self._mdib.process_incoming_alert_states_report(received_message_data)

    def _on_operational_state_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_operational_state_report')
        self._mdib.process_incoming_operational_state_report(received_message_data)

    def _on_waveform_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_waveform_report')
        self._mdib.process_incoming_waveform_states(received_message_data)

    def _on_episodic_context_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_episodic_context_report')
        self._mdib.process_incoming_context_report(received_message_data)

    def _on_episodic_component_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_episodic_component_report')
        self._mdib.process_incoming_component_report(received_message_data)

    def _on_description_modification_report(self, received_message_data: ReceivedMessage):
        self._logger.debug('_on_description_modification_report')
        self._mdib.process_incoming_description_modification_report(received_message_data)





"""The module implements the function mk_scopes."""

import urllib.parse

from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
from sdc11073.xml_types.wsd_types import ScopesType

# from IEEE Std 11073-20701-2018 chapter 9.3 SDC PARTICIPANT KEY PURPOSE based discovery
KEY_PURPOSE_SERVICE_PROVIDER = 'sdc.mds.pkp:1.2.840.10004.20701.1.1'
# from IEEE Std 11073-20701-2018 chapter 9.4 Context-based discovery
BICEPS_URI_UNK = 'biceps.uri.unk'


def _query_from_location_state(state) -> str:  # noqa: ANN001  typing is unknown here as it depends on the data_model in the mdib
    """Return a query string from a LocationContextStateContainer from GLUE 9.4.1.2."""
    query_dict: dict[str, str] = {}
    if state.LocationDetail is None:
        return ''
    if state.LocationDetail.Facility is not None:
        query_dict['fac'] = state.LocationDetail.Facility
    if state.LocationDetail.Building is not None:
        query_dict['bldng'] = state.LocationDetail.Building
    if state.LocationDetail.Floor is not None:
        query_dict['flr'] = state.LocationDetail.Floor
    if state.LocationDetail.PoC is not None:
        query_dict['poc'] = state.LocationDetail.PoC
    if state.LocationDetail.Room is not None:
        query_dict['rm'] = state.LocationDetail.Room
    if state.LocationDetail.Bed is not None:
        query_dict['bed'] = state.LocationDetail.Bed
    return urllib.parse.urlencode(query_dict, quote_via=urllib.parse.quote, safe='')


def mk_scopes(mdib: ProviderMdibProtocol) -> ScopesType:
    """Return a ScopesType instance.

    This method creates the scopes for publishing in wsdiscovery.
    """
    pm_types = mdib.data_model.pm_types
    pm_names = mdib.data_model.pm_names
    scope = ScopesType()
    for nodetype, scheme in (
        (pm_names.LocationContextDescriptor, 'sdc.ctxt.loc'),
        (pm_names.OperatorContextDescriptor, 'sdc.ctxt.opr'),
        (pm_names.EnsembleContextDescriptor, 'sdc.ctxt.ens'),
        (pm_names.WorkflowContextDescriptor, 'sdc.ctxt.wfl'),
        (pm_names.MeansContextDescriptor, 'sdc.ctxt.mns'),
    ):
        entities = mdib.entities.by_node_type(nodetype)
        for entity in entities:
            for state in [
                s for s in entity.states.values() if s.ContextAssociation == pm_types.ContextAssociation.ASSOCIATED
            ]:
                if not state.Identification:
                    msg = f'State {state.Handle} of type {nodetype} has no Identification element'
                    raise ValueError(msg)
                for ident in state.Identification:
                    # IEEE Std 11073-20701-2018 9.4 context based discovery
                    instance_identifier = (
                        f'/{urllib.parse.quote(ident.Root if ident.Root is not None else BICEPS_URI_UNK, safe="")}'
                    )
                    if ident.Extension:
                        instance_identifier += f'/{urllib.parse.quote(ident.Extension, safe="")}'
                    context_uri = f'{scheme}:{instance_identifier}'
                    query = ''
                    if nodetype == pm_names.LocationContextDescriptor:
                        if not state.LocationDetail:
                            msg = f'State {state.Handle} of type {nodetype} has no LocationDetail element'
                            raise ValueError(msg)
                        query = _query_from_location_state(state)
                    if query:
                        context_uri = f'{context_uri}?{query}'
                    scope.text.append(context_uri)

    scope.text.extend(_get_device_component_based_scopes(mdib))
    scope.text.append(KEY_PURPOSE_SERVICE_PROVIDER)  # default scope that is always included
    return scope


def _get_device_component_based_scopes(mdib: ProviderMdibProtocol) -> set[str]:
    """Return a set of scope strings.

    Use only MDSDescriptor, because there might be alot of VmdDescriptor that might exceed the dpws message size limit.
    Also, VmdDescriptor do not contain relevant information for discovery purposes.

    :return: a set of scope strings
    """
    pm_types = mdib.data_model.pm_types
    pm_names = mdib.data_model.pm_names
    scopes = set()
    entities = mdib.entities.by_node_type(pm_names.MdsDescriptor)
    for entity in entities:
        if entity.descriptor.Type is not None:
            coding_systems = (
                ''
                if entity.descriptor.Type.CodingSystem == pm_types.DEFAULT_CODING_SYSTEM
                else entity.descriptor.Type.CodingSystem
            )
            csv = entity.descriptor.Type.CodingSystemVersion or ''
            scope_string = f'sdc.cdc.type:/{coding_systems}/{csv}/{entity.descriptor.Type.Code}'
            scopes.add(scope_string)
    return scopes



"""Declare protocols for a Product and a WaveformProvider."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:

    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.provider.operations import OperationDefinitionBase
    from sdc11073.provider.sco import AbstractScoOperationsRegistry
    from sdc11073.xml_types.pm_types import ComponentActivation

    from .providerbase import OperationClassGetter
    from .waveformprovider.realtimesamples import AnnotatorProtocol
    from .waveformprovider.waveforms import WaveformGeneratorBase


class ProductProtocol:
    """A Product aggregates multiple role providers.

    The SdcProvider expects this interface.
    """

    def __init__(self,
                 mdib: ProviderMdibProtocol,
                 sco: AbstractScoOperationsRegistry,
                 log_prefix: str | None = None):
        """Create a product."""

    def init_operations(self):
        """Register all actively provided operations."""

    def stop(self):
        """Stop all role providers."""

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Call make_operation_instance of all role providers, until the first returns not None."""


class WaveformProviderProtocol(Protocol):
    """WaveformProvider is not a role provider of a product, it is separate.

    The SdcProvider expects this interface.
    """

    is_running: bool

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        ...

    def register_waveform_generator(self, descriptor_handle: str, wf_generator: WaveformGeneratorBase):
        """Add waveform generator to waveform sources."""
        ...

    def add_annotation_generator(self,
                                 coded_value: Any,
                                 trigger_handle: str,
                                 annotated_handles: list[str]) -> AnnotatorProtocol:
        """Add annotator to list of annotators."""
        ...

    def start(self):
        """Start whatever needs to be started in implementation."""
        ...

    def stop(self):
        """Stop whatever needs to be stopped in implementation."""
        ...

    def set_activation_state(self, descriptor_handle: str, component_activation_state: ComponentActivation):
        """Set the activation state of waveform generator and of Metric state in mdib."""

    def update_all_realtime_samples(self):
        """Update all realtime sample states that have a waveform generator registered."""



"""Communication logger."""
from __future__ import annotations

import functools
import logging
import pathlib
import threading
import time
from typing import Any, Callable

DISCOVERY_IN = 'sdc_comm.discovery.in'
DISCOVERY_OUT = 'sdc_comm.discovery.out'
SOAP_REQUEST_IN = 'sdc_comm.soap.request.in'
SOAP_REQUEST_OUT = 'sdc_comm.soap.request.out'
SOAP_RESPONSE_IN = 'sdc_comm.soap.response.in'
SOAP_RESPONSE_OUT = 'sdc_comm.soap.response.out'
SOAP_SUBSCRIPTION_IN = 'sdc_comm.soap.subscription.in'
WSDL = 'sdc_comm.wsdl'

LOGGER_NAMES = (DISCOVERY_IN, DISCOVERY_OUT, SOAP_REQUEST_IN, SOAP_REQUEST_OUT,
                SOAP_RESPONSE_IN, SOAP_RESPONSE_OUT, SOAP_SUBSCRIPTION_IN, WSDL)


class IpFilter(logging.Filter):
    """Filter all messages for the specific ip address."""

    def __init__(self, ip: str):
        super().__init__()
        self.ip = ip

    def filter(self, record: logging.LogRecord) -> bool:  # noqa: D102
        return self.ip == getattr(record, 'ip_address', None)


class CommLogger:
    """Base class to make configuring comm logger easy."""

    def __init__(self):
        self.handlers: dict[str, logging.Handler] = {}

    def start(self) -> None:
        """Start logger."""
        for name, handler in self.handlers.items():
            logger = logging.getLogger(name)
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)

    def stop(self) -> None:
        """Stop logger."""
        for name, handler in self.handlers.items():
            logger = logging.getLogger(name)
            logger.removeHandler(handler)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args, **kwargs):  # noqa: ANN002, ANN003
        self.stop()


class DirectoryLogger(CommLogger):
    """Logger writing communication logs into a directory. Each message will be contained in a single file."""

    D_IN = 'in'
    D_OUT = 'out'

    T_UDP = 'udp'
    T_UDP_MULTICAST = 'udpB'
    T_HTTP = 'http_subscr'
    T_WSDL = 'wsdl'

    T_HTTP_REQ = 'http_req'
    T_HTTP_RESP = 'http_resp'

    def __init__(self, log_folder: str | pathlib.Path, log_out: bool = False, log_in: bool = False,
                 broadcast_ip_filter: str | None = None):
        super().__init__()
        self._log_folder = pathlib.Path(log_folder)
        self._counter = 1
        self._io_lock = threading.Lock()

        if log_in:
            self.handlers.update({
                DISCOVERY_IN: self._GenericHandler(functools.partial(self._write_log, self.T_UDP, self.D_IN)),
                SOAP_REQUEST_IN: self._GenericHandler(functools.partial(self._write_log, self.T_HTTP_REQ, self.D_IN)),
                SOAP_RESPONSE_IN: self._GenericHandler(functools.partial(self._write_log, self.T_HTTP_RESP, self.D_IN)),
                SOAP_SUBSCRIPTION_IN: self._GenericHandler(functools.partial(self._write_log, self.T_HTTP, self.D_IN)),
                WSDL: self._GenericHandler(functools.partial(self._write_log, self.T_WSDL, self.D_IN)),
            })
        if log_out:
            self.handlers.update({
                DISCOVERY_OUT: self._GenericHandler(functools.partial(self._write_log, self.T_UDP, self.D_OUT)),
                SOAP_REQUEST_OUT: self._GenericHandler(functools.partial(self._write_log, self.T_HTTP_REQ, self.D_OUT)),
                SOAP_RESPONSE_OUT: self._GenericHandler(
                    functools.partial(self._write_log, self.T_HTTP_RESP, self.D_OUT)),
            })
        if broadcast_ip_filter:
            broadcast_filter = IpFilter(broadcast_ip_filter)
            if DISCOVERY_IN in self.handlers:
                self.handlers[DISCOVERY_IN].addFilter(broadcast_filter)
            if DISCOVERY_OUT in self.handlers:
                self.handlers[DISCOVERY_OUT].addFilter(broadcast_filter)

    def start(self) -> None:
        self._log_folder.mkdir(parents=True, exist_ok=True)
        super().start()

    def _mk_filename(self, ip_type: str, direction: str, *infos: str) -> str:
        """Create file name.

        :param ip_type: "tcp" or "udp"
        :param direction: "in" or "out"
        :param info: becomes part of filename
        :return:
        """
        assert ip_type in (self.T_UDP, self.T_UDP_MULTICAST, self.T_HTTP,
                           self.T_HTTP_REQ, self.T_HTTP_RESP, self.T_WSDL)
        assert direction in (self.D_IN, self.D_OUT)
        extension = 'wsdl' if ip_type == self.T_WSDL else 'xml'
        time_string = f'{time.time():06.3f}'[-8:]
        self._counter += 1
        info_text = f'-{"-".join(infos)}' if infos else ''
        return f'{time_string}-{direction}-{ip_type}{info_text}.{extension}'

    def _write_log(self, ttype: str, direction: str, msg: bytes, *infos: str) -> None:
        path = self._log_folder.joinpath(self._mk_filename(ttype, direction, *infos))
        with self._io_lock:
            path.write_bytes(msg)

    class _GenericHandler(logging.Handler):
        def __init__(self, emit: Callable):
            super().__init__()
            self._emit = emit

        def emit(self, record: logging.LogRecord) -> None:
            try:
                msg = self.format(record).encode()  # defaults to utf-8
                args = []
                if ip := getattr(record, 'ip_address', None):
                    args.append(ip)
                if http_method := getattr(record, 'http_method', None):
                    args.append(http_method)
                self._emit(msg, *args)
            except Exception:  # noqa: BLE001
                self.handleError(record)


class StreamLogger(CommLogger):
    """Set a stream handler for each comm logger."""

    def __init__(self, stream: Any | None = None, broadcast_ip_filter: str | None = None):
        super().__init__()
        for name in LOGGER_NAMES:
            self.handlers[name] = self._get_handler(stream)

        if broadcast_ip_filter:
            broadcast_filter = IpFilter(broadcast_ip_filter)
            if DISCOVERY_IN in self.handlers:
                self.handlers[DISCOVERY_IN].addFilter(broadcast_filter)
            if DISCOVERY_OUT in self.handlers:
                self.handlers[DISCOVERY_OUT].addFilter(broadcast_filter)

    @staticmethod
    def _get_handler(stream: Any | None) -> logging.StreamHandler:
        handler = logging.StreamHandler(stream=stream)
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(message)s'))
        return handler
